from .archive import archive
from .recall import recall_dev, recall_single
from .retrieve import retrieve, retrieve_improved

__all__ = [
    "archive",
    "recall_dev",
    "recall_single",
    "retrieve",
    "retrieve_improved",
]



import inspect
import json
import time
import warnings
from pathlib import Path
from typing import Union

from ..base import FileGroup, GroupCollection, ls, search
from ..constants import PYSLK_SEARCH_DELAY, PYSLK_WILDCARDS
from ..pyslk_exceptions import PySlkEmptyInputError
from ..raw import group_files_by_tape_raw, retrieve_raw
from .gen_queries import gen_file_query
from .resource_extras import resource_exists
from .storage import is_cached

__all__ = [
    "_check_input_gfbt",
    "_gen_group_collection",
    "_group_retrieve",
    "count_tapes",
    "count_tapes_with_multi_tape_files",
    "count_tapes_with_single_tape_files",
    "group_files_by_tape",
]


def group_files_by_tape(
    resource_path: Union[Path, str, list, None] = None,
    resource_ids: Union[str, list, int, None] = None,
    search_id: Union[str, int, None] = None,
    search_query: Union[str, None] = None,
    recursive: bool = False,
    max_tape_number_per_search: int = -1,
    run_search_query: bool = False,
    evaluate_regex_in_input: bool = False,
) -> list[dict]:
    """Group files by tape id.

    Group a list of files by their tape id. Has not all arguments of the `slk_helpers group_files_by_tape` cli call.
    Please us :py:meth:`pyslk.count_tapes` to count the number of tapes onto which files are stored on.

    :param resource_path: list of files or a namespaces with files that should be grouped.
    :type resource_path: ``str``, ``list``, ``Path``
    :param resource_ids: list of resource ids to be searched for
    :type resource_ids: str or list or Path-like or None
    :param search_id: id of a search
    :type search_id: ``int``, ``str``
    :param search_query: a search query
    :type search_query: ``str``
    :param recursive: do recursive search in the namespaces
    :type recursive: ``bool``
    :param max_tape_number_per_search: number of tapes per search; if '-1' => the parameter is not set
    :type max_tape_number_per_search: ``int``
    :param run_search_query: generate and run (a) search query strings instead of the lists of files per tape and print
            the search i, Default: False
    :type run_search_query: ``bool``
    :param evaluate_regex_in_input: expect that a regular expression in part of the input (file names) and evalute it
    :type evaluate_regex_in_input: bool
    :return: A list of dictionaries containing group and tape info.
    :rtype: ``list[dict]``

    .. seealso::
        * :py:meth:`~pyslk.count_tapes`
        * :py:meth:`~pyslk.group_files_by_tape_raw`

    .. rubric:: Examples

    .. code-block:: python

        >>> import pyslk as slk
        >>> slk.group_files_by_tape(["/test/test3/ingest_01_102", "/test/test3/ingest_01_339"])
        [{'id': -1,
          'location': 'cache',
          'label': '',
          'status': '',
          'file_count': 2,
          'files': ['/test/test3/ingest_01_102', '/test/test3/ingest_01_339'],
          'search_query': '{"$and":[{"path":{"$gte":"/test/test3","$max_depth":1}},
                            {"resources.name":{"$regex":"ingest_01_102|ingest_01_339"}}]}',
          'search_id': 416837}]
    """
    # check if input values are OK
    _check_input_gfbt(
        calling_function=inspect.stack()[0][3],
        resource_path=resource_path,
        resource_ids=resource_ids,
        search_id=search_id,
        search_query=search_query,
        recursive=recursive,
    )
    # only grouping by tape id / barcode makes sense, otherwise we have duplicated keys.
    output = group_files_by_tape_raw(
        resource_paths=resource_path,
        resource_ids=resource_ids,
        search_id=search_id,
        search_query=search_query,
        print_tape_barcode=True,
        recursive=recursive,
        json=True,
        gen_search_query=True,
        set_max_tape_number_per_search=max_tape_number_per_search,
        run_search_query=run_search_query,
        evaluate_regex_in_input=evaluate_regex_in_input,
    )
    return json.loads(output)


def _group_retrieve(
    resource: Union[str, Path, int, list],
    dest_dir: Union[str, Path],
    recursive: bool,
    delayed: bool = False,
    **kwargs,
) -> list:
    """
    Grouped retrieve

    Helper function used by :py:meth:`pyslk.retrieve`. Uses :py:meth:`pyslk.group_files_by_tape`.

    :param resource: a resource (path to it or resource id)
    :type resource: ``str`` or ``Path`` or ``int`` or ``list``
    :param dest_dir: destination directory
    :type dest_dir: ``str`` or ``Path``
    :param recursive: do recursive retrieval
    :type recursive: ``bool``
    :param delayed: do delayed retrieval
    :type delayed: ``bool``
    :param `**kwargs`: keyword args of :py:meth:`pyslk.retrieve`
    :return: a list
    :rtype: ``list``

    .. seealso::
        * :py:meth:`~pyslk.retrieve`
        * :py:meth:`~pyslk.group_files_by_tape`
    """

    if delayed is True:
        from dask import delayed as delay
    else:

        def delay(x):
            return x

    # determine whether we have a search id or a resource
    #  whereas a resource might be
    #    * a list of resources,
    #    * an expression with wildcards or
    #    * an individual resource represented as str or Path-like obj
    search_id: Union[int, None] = None
    # str with wildcards
    if isinstance(resource, str) and PYSLK_WILDCARDS in resource:
        # create file list and group them
        resource = list(ls(resource).filename)
    elif isinstance(resource, Path):
        # resource is a path => do nothing except for checking whether it exists
        if not resource_exists(resource):
            warnings.warn(
                f"pyslk.{inspect.stack()[0][3]}: resource does not exist: {str(resource)}"
            )
    elif isinstance(resource, list):
        # resource is a list => all items need to be string or Path-like
        wrong_type: list = [
            type(res).__name__ for res in resource if not isinstance(res, (str, Path))
        ]
        if len(wrong_type) > 0:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: if 'resource' is a list then the list's items have to be of type "
                + f"'str' or Path-like; {len(wrong_type)} items are affected having the types "
                + f"{', '.join(set(wrong_type))}"
            )
        # check if all items of list exist
        not_exist: list = [res for res in resource if not resource_exists(res)]
        if len(not_exist) > 0:
            warnings.warn(
                f"pyslk.{inspect.stack()[0][3]}: some resources does not exist: "
                + ", ".join(not_exist)
            )
    elif isinstance(resource, int):
        # resource is a search id
        search_id = resource
        resource = None
    elif isinstance(resource, str):
        # resource can either be a resource or a search id
        try:
            search_id = int(resource)
            resource = None
        except ValueError:
            pass
        if resource is not None:
            # resource is a path => do nothing except for checking whether it exists
            if not resource_exists(resource):
                warnings.warn(
                    f"pyslk.{inspect.stack()[0][3]}: resource does not exist: "
                    + resource
                )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' has wrong type; need 'str', 'int' or 'Path' but "
            + f"got {type(resource).__name__}"
        )

    groups = group_files_by_tape(
        resource_path=resource,
        search_id=search_id,
        recursive=recursive,
        run_search_query=True,
        evaluate_regex_in_input=True,
    )

    # HERE WE HAVE TO WAIT?! Otherwise, the search_ids from the grouped
    # search queries will not be valid for a retrieve.
    if delayed is False:
        time.sleep(PYSLK_SEARCH_DELAY)

    # return grouped retrieves
    return [
        delay(retrieve_raw)(g["search_id"], dest_dir, recursive=recursive, **kwargs)
        for g in groups
    ]


def _gen_group_collection(
    resource_path: Union[str, Path, list[str], None] = None,
    resource_ids: Union[str, list, int, None] = None,
    search_id: Union[str, int, None] = None,
    search_query: Union[str, None] = None,
    recursive: bool = False,
    max_tape_number_per_search: int = -1,
    split_recalls: bool = True,
    verbose: bool = False,
    evaluate_regex_in_input: bool = False,
) -> GroupCollection:
    """

    output if not details:
        {int: int}

        key: job_id
        value: job_status: -1 => running/queued; 0 => finished; >0 => error

    :param resource_path: list of files or a namespaces with files that should be recalled.
    :type resource_path: ``str``, ``list``, ``Path``
    :param resource_ids: list of resource ids to be searched for
    :type resource_ids: str or list or Path-like or None
    :param search_id: id of a search
    :type search_id: ``int``, ``str``
    :param search_query: a search query
    :type search_query: ``str``
    :param recursive: do recursive search in the namespaces
    :type recursive: ``bool``
    :param max_tape_number_per_search: number of tapes per search; if '-1' => the parameter is not set
    :type max_tape_number_per_search: ``int``
    :param split_recalls: run one recall per tape ...
    :type split_recalls: ``bool``
    :param verbose: verbose mode
    :type verbose: ``bool``
    :param evaluate_regex_in_input: expect that a regular expression in part of the input (file names) and evalute it
    :type evaluate_regex_in_input: bool
    :return:
    """
    # check if input values are OK
    _check_input_gfbt(
        calling_function=inspect.stack()[0][3],
        resource_path=resource_path,
        resource_ids=resource_ids,
        search_id=search_id,
        search_query=search_query,
        recursive=recursive,
    )

    # TODO: further implement 'verbose'

    if not isinstance(verbose, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'verbose' has wrong type; need 'bool' but "
            + f"got {type(verbose).__name__}"
        )

    output: GroupCollection = GroupCollection()
    tape_group: dict

    if split_recalls:
        if verbose:
            print(f"pyslk.{inspect.stack()[0][3]}: split recalls")
        # ~~~~~~~~~~~~~~~~ SPLIT RECALL AND ... ~~~~~~~~~~~~~~~~
        tape_groups = group_files_by_tape(
            resource_path=resource_path,
            resource_ids=resource_ids,
            search_id=search_id,
            search_query=search_query,
            recursive=recursive,
            run_search_query=True,
            max_tape_number_per_search=max_tape_number_per_search,
            evaluate_regex_in_input=evaluate_regex_in_input,
        )
        if verbose:
            print(f"pyslk.{inspect.stack()[0][3]}: splitting recalls successful")
        if len(tape_groups) == 0:
            # ~~~~~~~~~~~~~~~~ ... NO GROUP ~~~~~~~~~~~~~~~~
            raise PySlkEmptyInputError(
                f"pyslk.{inspect.stack()[0][3]}: No content to recall"
            )
            # => error
        # consider each tape group as a job and add it to the JobCollection
        if verbose:
            print(
                f"pyslk.{inspect.stack()[0][3]}: request split into {len(tape_groups)} recalls"
            )
        for tape_group in tape_groups:
            output.add_group(FileGroup(tape_group))
    else:
        if verbose:
            print(f"pyslk.{inspect.stack()[0][3]}: not split recalls")
        tape_group = dict()
        # we need the search_id
        # * if search_query or resource_path are not None, search_id will be None
        # * if search_query and resource_path are None, search_id will not be None
        if search_query is not None:
            if verbose:
                print(
                    f"pyslk.{inspect.stack()[0][3]}: run search for provided search query"
                )
            search_id = search(search_query)
        elif resource_path is not None:
            if verbose:
                print(
                    f"pyslk.{inspect.stack()[0][3]}: run search for provided resource path"
                )
            # => run search ...
            search_query = gen_file_query(resource_path, recursive)
            search_id = search(search_query)
        if verbose:
            print(
                f"pyslk.{inspect.stack()[0][3]}: collect files from search id {search_id}"
            )
        # get file list
        files: list[str] = list(ls(search_id).filename)
        if verbose:
            print(f"pyslk.{inspect.stack()[0][3]}: files collected: {files}")
        # set values in group dict
        tape_group["search_query"] = search_query
        tape_group["search_id"] = search_id
        tape_group["files"] = files
        tape_group["file_count"] = len(files)
        if is_cached(search_id=search_id):
            tape_group["location"] = "cache"
        else:
            tape_group["location"] = "tape"
        # generate a FileGroup from this
        output.add_group(FileGroup(tape_group))
    # return GroupCollection
    return output


def _check_input_gfbt(
    calling_function: str,
    resource_path: Union[str, list, Path, None] = None,
    resource_ids: Union[str, list, int, None] = None,
    search_id: Union[str, int, None] = None,
    search_query: Union[str, None] = None,
    recursive: bool = False,
):
    """
    Check if input values are correct; if not, throw errors

    Thrown errors

    * PySlkException
    * TypeError
    * ValueError

    :param calling_function: name of the functions which calls this function
    :type calling_function: ``str``
    :param resource_path: a resource path (str or Path) or multiple resource paths (in a list)
    :type resource_path: ``str`` or ``path-like`` or ``list``
    :param resource_ids: list of resource ids to be searched for
    :type resource_ids: str or list or Path-like or None
    :param search_id: id of a search
    :type search_id: ``int``, ``str``
    :param search_query: a search query
    :type search_query: ``str``
    :param recursive: set whether resource should be evaluated recursively or not
    :type recursive: ``bool``
    """
    # check consistency of input
    if (
        (resource_path is not None and search_id is not None)
        or (resource_path is not None and search_query is not None)
        or (search_id is not None and search_query is not None)
        or (resource_path is not None and resource_ids is not None)
        or (search_id is not None and resource_ids is not None)
        or (search_query is not None and resource_ids is not None)
    ):
        raise ValueError(
            f"pyslk.{calling_function}: only resource_path xor search_id xor search_query can be set (== only one)"
        )
    if (
        resource_path is None
        and search_id is None
        and search_query is None
        and resource_ids is None
    ):
        raise ValueError(
            f"pyslk.{calling_function}: either resource_path or search_id or search_query have to be set"
        )
    # make type checks: resource_path
    if resource_path is not None:
        if not isinstance(resource_path, (str, Path, list)):
            raise TypeError(
                f"pyslk.{calling_function}: wrong input type; need 'str', 'Path' or 'list'; got "
                + f"'{type(resource_path).__name__}'"
            )
        if isinstance(resource_path, (str, Path)):
            if not resource_exists(resource_path):
                raise FileNotFoundError(
                    f"pyslk.{calling_function}: resource does not exist {resource_path}"
                )
        else:
            # resource_path is a list
            non_exist: list = [rp for rp in resource_path if not resource_exists(rp)]
            if len(non_exist) > 0:
                raise FileNotFoundError(
                    f"pyslk.{calling_function}: one or more resources do not exist {', '.join(non_exist)}"
                )
    if search_id is not None:
        if not isinstance(search_id, (str, int)):
            raise TypeError(
                f"pyslk.{calling_function}: wrong input type; need 'int' or 'str'; got '{type(search_id).__name__}'"
            )
        if isinstance(search_id, str):
            try:
                search_id = int(search_id)
            except ValueError:
                raise ValueError(
                    f"pyslk.{calling_function}: got 'str' as input for 'search_id' but cannot convert it to "
                    + f"'int'; value is '{search_id}'"
                )
    if search_query is not None:
        if not isinstance(search_query, str):
            raise TypeError(
                f"pyslk.{calling_function}: wrong input type; need 'str'; got '{type(search_query).__name__}'"
            )
    if not isinstance(recursive, bool):
        raise TypeError(
            f"pyslk.{calling_function}: wrong input type; need 'bool'; got '{type(recursive).__name__}'"
        )
    # make type checks: resource_ids
    if resource_ids is not None:
        if not isinstance(resource_ids, (str, int, list)):
            raise TypeError(
                f"pyslk.{calling_function}: wrong input type; need 'str', 'int' or 'list'; got "
                + f"'{type(resource_ids).__name__}'"
            )
        if isinstance(resource_ids, (str, int)):
            if not resource_exists(resource_ids):
                raise FileNotFoundError(
                    f"pyslk.{calling_function}: resource does not exist {resource_ids}"
                )
        else:
            # resource_ids are a list
            non_exist: list = [rp for rp in resource_ids if not resource_exists(rp)]
            if len(non_exist) > 0:
                raise FileNotFoundError(
                    f"pyslk.{calling_function}: one or more resources do not exist {', '.join(non_exist)}"
                )


def count_tapes(
    resource_path: Union[str, list, Path, None] = None,
    resource_ids: Union[str, list, int, None] = None,
    search_id: Union[str, int, None] = None,
    search_query: Union[str, None] = None,
    recursive: bool = False,
    evaluate_regex_in_input: bool = False,
) -> dict[str, int]:
    """
    Count number of tapes onto which provided files are stored; distinguishes between multi-tape and single-tape files

    :param resource_path: a resource path (str or Path) or multiple resource paths (in a list)
    :type resource_path: ``str`` or ``path-like`` or ``list``
    :param resource_ids: list of resource ids to be searched for
    :type resource_ids: str or list or Path-like or None
    :param search_id: id of a search
    :type search_id: ``int``, ``str``
    :param search_query: a search query
    :type search_query: ``str``
    :param recursive: set whether resource should be evaluated recursively or not
    :type recursive: ``bool``
    :param evaluate_regex_in_input: expect that a regular expression in part of the input (file names) and evalute it
    :type evaluate_regex_in_input: bool
    :return: dictionary containing the two tape counts
    :rtype: ``dict``

    .. seealso::
        * :py:meth:`~pyslk.group_files_by_tape`
        * :py:meth:`~pyslk.group_files_by_tape_raw`
    """
    # check if input values are OK
    _check_input_gfbt(
        calling_function=inspect.stack()[0][3],
        resource_path=resource_path,
        resource_ids=resource_ids,
        search_id=search_id,
        search_query=search_query,
        recursive=recursive,
    )
    output: dict = dict()
    # run group_files_by_tape
    raw_output = group_files_by_tape_raw(
        resource_paths=resource_path,
        resource_ids=resource_ids,
        search_id=search_id,
        search_query=search_query,
        recursive=recursive,
        count_tapes=True,
        evaluate_regex_in_input=evaluate_regex_in_input,
    ).split("\n")
    # look for output line with single-tape-files
    for i in raw_output:
        if "single-tape" in i:
            output["n_tapes__single_tape_files"] = int(i.split(" ")[0])
            break
    # look for output line with multi-tape-files
    for i in raw_output:
        if "multi-tape" in i:
            output["n_tapes__multi_tape_files"] = int(i.split(" ")[0])
            break
    # return dict
    return output


def count_tapes_with_single_tape_files(
    resource_path: Union[str, int, Path, None] = None,
    resource_ids: Union[str, list, int, None] = None,
    search_id: Union[str, int, None] = None,
    search_query: Union[str, None] = None,
) -> int:
    """
    Count number of tapes onto which provided files are stored which are not split onto multiple tapes per file

    Internally calls :py:meth:`pyslk.count_tapes`

    :param resource_path: a resource path
    :type resource_path: ``str`` or ``int`` or ``path-like``
    :param resource_ids: list of resource ids to be searched for
    :type resource_ids: str or list or Path-like or None
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :param search_query: a search query
    :type search_query: ``str``
    :return: number of tapes
    :rtype: ``int``

    .. seealso::
        * :py:meth:`~pyslk.count_tapes`
        * :py:meth:`~pyslk.count_tapes_with_multi_tape_files`
    """
    return count_tapes(
        resource_path=resource_path,
        resource_ids=resource_ids,
        search_id=search_id,
        search_query=search_query,
    )["n_tapes__single_tape_files"]


def count_tapes_with_multi_tape_files(
    resource_path: Union[str, int, Path, None] = None,
    resource_ids: Union[str, list, int, None] = None,
    search_id: Union[str, int, None] = None,
    search_query: Union[str, None] = None,
) -> int:
    """
    Count number of tapes onto which provided files are stored which are split onto multiple tapes per file

    Internally calls :py:meth:`pyslk.count_tapes`

    :param resource_path: a resource path
    :type resource_path: ``str`` or ``int`` or ``path-like``
    :param resource_ids: list of resource ids to be searched for
    :type resource_ids: str or list or Path-like or None
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :param search_query: a search query
    :type search_query: ``str``
    :return: number of tapes
    :rtype: ``int``

    .. seealso::
        * :py:meth:`~pyslk.count_tapes`
        * :py:meth:`~pyslk.count_tapes_with_single_tape_files`
    """
    return count_tapes(
        resource_path=resource_path,
        resource_ids=resource_ids,
        search_id=search_id,
        search_query=search_query,
    )["n_tapes__multi_tape_files"]



from __future__ import annotations

import inspect
import json
from pathlib import Path
from typing import Union

from ..pyslk_exceptions import PySlkException
from ..raw import iscached_raw, job_status_raw
from .stati import StatusJob

__all__: list = ["FileGroup"]


class FileGroup:
    def __init__(self, group_dict: dict):
        # tape_id
        self.tape_ids: list[int] = list()
        tmp_tape_id = group_dict.get("tape_id", None)
        if tmp_tape_id is not None:
            self.tape_ids.append(tmp_tape_id)
        # tape_barcode
        self.tape_barcodes: list[str] = list()
        tmp_tape_barcode = group_dict.get("tape_barcode", None)
        if tmp_tape_barcode is not None:
            self.tape_barcodes.append(tmp_tape_barcode)
        # job_id
        self.job_ids: list[int] = list()
        tmp_job_id: Union[int, None] = group_dict.get("job_id", None)
        if tmp_job_id is not None:
            if isinstance(tmp_job_id, list):
                self.job_ids.extend(tmp_job_id)
            else:
                self.job_ids.append(tmp_job_id)
        # process_id
        self.process_ids: list[int] = list()
        tmp_process_id = group_dict.get("process_id", group_dict.get("pid", None))
        if tmp_process_id is not None:
            if isinstance(tmp_process_id, list):
                self.process_ids.extend(tmp_process_id)
            else:
                self.process_ids.append(tmp_process_id)
        # search_id
        self.search_id: int = group_dict.get("search_id", -1)
        # search_query
        self.search_query: str = group_dict.get("search_query", None)
        # files
        self.files: Union[list[str], list[Path]] = group_dict.get("files", list())
        # past caching status
        self.cached: bool = group_dict.get("location", "") == "cache"
        # initialize dict for job stati
        self.job_stati: dict = dict()
        # set error flag for failed recall jobs
        self.recall_error: bool = False

    def get_job_status(self) -> StatusJob:
        if len(self.job_ids) == 0:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: tried to get job status although no job id is set"
            )
        job_id: int = self.job_ids[-1]
        job_status: StatusJob = self.job_stati.get(job_id, None)
        # IF: no status exists for this job id or the job status indicates that an updated might be needed
        if job_status is None or job_status not in [
            StatusJob.COMPLETED,
            StatusJob.SUCCESSFUL,
            StatusJob.FAILED,
            StatusJob.ABORTED,
        ]:
            # run raw job_status function
            output = job_status_raw(job_id, return_type=2)
            if output.returncode in [0, 1]:
                # convert it to a proper status
                job_status = StatusJob(output.stdout.decode("utf-8")[:-1])
                self.job_stati[job_id] = job_status
            else:
                raise PySlkException(
                    f"pyslk.{inspect.stack()[0][3]}: "
                    + output.stdout.decode("utf-8")
                    + " "
                    + output.stderr.decode("utf-8")
                )

        return job_status

    def get_tape_ids(self) -> list[int]:
        return self.tape_ids

    def get_tape_barcodes(self) -> list[str]:
        return self.tape_barcodes

    def get_search_id(self) -> int:
        return self.search_id

    def get_search_query(self) -> str:
        return self.search_query

    def get_files(self) -> Union[list[str], list[Path]]:
        return self.files

    def is_cached(self) -> bool:
        return self.cached

    def updated_caching_info(self) -> bool:
        """
        Check is all files of this job are cached. Throws an error if this job is associated to no file.

        :return: info if all files of this job are cached
        :rtype: bool
        """
        if self.search_id != -1:
            self.cached = _cached_duplicate(search_id=self.search_id)
        elif self.files is not None and len(self.files) > 0:
            self.cached = all([_cached_duplicate(f) for f in self.files])
        else:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: neither search id ('search_id') nor file list ('files')"
            )
        return self.cached

    def get_job_id(self) -> int:
        """
        Returns the current job id of this job

        :return: job id
        :rtype: int
        """
        if len(self.job_ids) == 0:
            return -1
        return self.job_ids[-1]

    def get_job_ids(self) -> list[int]:
        """
        Prints the history of job IDs of this job. The first one is the oldest and the last one the newest.

        :return: all job ids which have been associated to this job
        :rtype: list[int]
        """
        return self.job_ids

    def put_job_id(self, job_id: int):
        if len(self.job_ids) == 0 or job_id != self.job_ids[-1]:
            self.job_ids.append(job_id)

    def get_process_id(self) -> int:
        """
        Returns the process id slk call which submitted the last recall job

        :return: process id
        :rtype: int
        """
        if len(self.process_ids) == 0:
            return -1
        return self.process_ids[-1]

    def get_pid(self) -> int:
        """
        Same as get_process_id(). Returns the process id slk call which submitted the last recall job

        :return: process id
        :rtype: int
        """
        return self.get_process_id()

    def get_process_ids(self) -> list[int]:
        """
        Returns the history of the process ids of the slk calls which submitted the recall jobs

        :return: all process ids which have been associated to this group
        :rtype: list[int]
        """
        return self.process_ids

    def get_pids(self) -> list[int]:
        """
        Same as get_process_ids(). Returns the history of the process ids of the slk calls which submitted the recalls

        :return: all process ids which have been associated to this group
        :rtype: list[int]
        """
        return self.get_process_ids()

    def put_process_id(self, process_id: int):
        """
        Set the current pid / process id of slk call which submitted the current recall job

        :param process_id: process id / pid to be set
        :type process_id: int
        """
        if len(self.process_ids) == 0 or process_id != self.process_ids[-1]:
            self.process_ids.append(process_id)

    def put_pid(self, process_id: int):
        """
        Same as put_process_id(). Set the current pid / process id of slk call which submitted the current recall job

        :param process_id: process id / pid to be set
        :type process_id: int
        """
        self.put_process_id(process_id)

    def is_recall_error(self) -> bool:
        return self.recall_error

    def set_recall_error(self, recall_error):
        self.recall_error = recall_error

    def has_tape_ids(self) -> bool:
        return len(self.tape_ids) > 0

    def has_tape_barcodes(self) -> bool:
        return len(self.tape_barcodes) > 0

    def has_job_id(self) -> bool:
        return len(self.job_ids) > 0

    def has_process_id(self) -> bool:
        return len(self.process_ids) > 0

    def has_pid(self) -> bool:
        return self.has_process_id()

    def has_search_id(self) -> bool:
        return self.search_id != -1

    def has_search_query(self) -> bool:
        return self.search_query is not None

    def has_files(self) -> bool:
        return len(self.files) > 0

    def put_tape_id(self, tape_id: int):
        if tape_id not in self.tape_ids:
            self.tape_ids.append(tape_id)

    def put_tape_barcode(self, tape_barcode: str):
        if tape_barcode not in self.tape_barcodes:
            self.tape_barcodes.append(tape_barcode)

    def put_files(self, files: Union[list[str], list[Path]]):
        for f in files:
            if f not in self.files:
                self.files.append(f)

    def merge_group(self, group: FileGroup):
        if self.has_search_id() and group.has_search_id():
            if self.search_id != group.get_search_id():
                raise PySlkException(
                    f"pyslk.{inspect.stack()[0][3]}: search ids do not agree. this search id: {self.search_id}. "
                    + f"other search id: {group.get_search_id()}"
                )
        for tape_id in group.get_tape_ids():
            self.put_tape_id(tape_id)
        for tape_barcode in group.get_tape_barcodes():
            self.put_tape_barcode(tape_barcode)
        self.put_files(group.get_files())

    def dump(self) -> dict:
        """
        Dump all details  of this job as a dictionary

        :return: a dictionary of all job details
        :rtype: dict
        """
        tmp_dict = dict()
        if self.has_job_id():
            tmp_dict["job_id"] = self.get_job_id()
            tmp_dict["job_ids"] = self.get_job_ids()
            tmp_dict["status"] = self.get_job_status()
        if self.has_tape_ids():
            tmp_dict["tape_ids"] = self.get_tape_ids()
        if self.has_tape_barcodes():
            tmp_dict["tape_barcodes"] = self.get_tape_barcodes()
        if self.has_search_id():
            tmp_dict["search_id"] = self.get_search_id()
        if self.has_search_query():
            tmp_dict["search_query"] = json.loads(self.get_search_query())
        if self.has_files():
            tmp_dict["files"] = self.get_files()
        tmp_dict["cached"] = self.is_cached()
        return tmp_dict

    def __str__(self) -> str:
        """
        print a JSON string of all details of this job

        :return: JSON strong of all details of this job
        :rtype: str
        """
        tmp_dict: dict = self.dump()
        if self.has_job_id():
            tmp_dict["status"] = self.get_job_status().__str__()
        return json.dumps(tmp_dict)

    def json(self) -> str:
        """
        print a JSON string of all details of this job

        :return: JSON string of all details of this job
        :rtype: str
        """
        return self.__str__()


def _cached_duplicate(
    resource: Union[str, Path, None] = None,
    search_id: Union[int, None] = None,
) -> Union[bool, dict[str, list[Path]]]:
    """Check if whether file(s) is/are in HSM cache or not; returns True/False or a dict with keys 'cached' and
        'not_cached', depending on 'details'

    :param resource: a resource path
    :type resource: str or path-like
    :param search_id: id of a search
    :type search_id: int
    :returns: True if file is in cache; False otherwise; if 'details': dictionary with an entry per file
    :rtype: bool or dict
    """
    output = iscached_raw(
        resource_path=resource, search_id=search_id, recursive=True, return_type=2
    )
    if output.returncode == 0:
        return True
    elif output.returncode == 1:
        return False
    else:
        # output.returncode is neither 0 nor 1
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + output.stdout.decode("utf-8")
            + " "
            + output.stderr.decode("utf-8")
        )



import inspect
from typing import Union

from ..raw import version_slk_helpers_raw, version_slk_raw

__all__ = ["cli_versions", "version_slk", "version_slk_helpers"]


def cli_versions(cli: Union[list[str], tuple[str], str, None] = None) -> dict[str, str]:
    """Return the version of slk and/or slk_helpers

    Uses :py:meth:`~pyslk.version_slk_raw` and :py:meth:`~pyslk.version_slk_helpers_raw`

    Similar to `~pyslk.version_slk` and `~pyslk.version_slk_helpers`

    :param cli: names of the commands for which the version should be obtained
    :type cli: ``list[str]``, tuple[str], ``str``
    :return: dictionary containing the versions of clis slk and/or slk_helpers
    :rtype: ``dict``

    .. seealso::
        * :py:meth:`~pyslk.version_slk`
        * :py:meth:`~pyslk.version_slk_helpers`
        * :py:meth:`~pyslk.version_slk_helpers_raw`
        * :py:meth:`~pyslk.version_slk_raw`

    """
    # check type
    if cli is not None and not isinstance(cli, (list, tuple, str)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'cli' has wrong type; need 'list[str]', 'tuple[str]', 'str' or 'None' but "
            + f"got '{type(cli).__name__}'"
        )
    # define output dict
    output = dict()
    if (
        cli is None
        or (isinstance(cli, (list, tuple)) and "slk" in cli)
        or (isinstance(cli, str) and "slk" == cli)
    ):
        output["slk"] = version_slk_raw()
    if (
        cli is None
        or (isinstance(cli, (list, tuple)) and "slk_helpers" in cli)
        or (isinstance(cli, str) and "slk_helpers" == cli)
    ):
        output["slk_helpers"] = version_slk_helpers_raw()
    return output


def version_slk() -> str:
    """List the version of slk

    Uses :py:meth:`~pyslk.version_slk_raw`

    :returns: stdout of the slk call
    :rtype: ``str``

    .. seealso::
        * :py:meth:`~pyslk.cli_versions`
        * :py:meth:`~pyslk.version_slk`
        * :py:meth:`~pyslk.version_slk_raw`
    """
    return version_slk_raw()


def version_slk_helpers() -> str:
    """List the version of slk_helpers

    Uses :py:meth:`~pyslk.version_slk_helpers_raw`

    :returns: stdout of the slk_helpers call
    :rtype: ``str``

    .. seealso::
        * :py:meth:`~pyslk.cli_versions`
        * :py:meth:`~pyslk.version_slk_helpers`
        * :py:meth:`~pyslk.version_slk_helpers_raw`
    """
    return version_slk_helpers_raw()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""\
exceptions provides the class PySlkException, which is the general
exception for the pyslk package.
"""

__all__ = [
    "PySlkException",
    "PySlkNoValidLoginTokenError",
    "PySlkBadFileError",
    "PySlkBadProcessError",
    "ArchiveError",
    "PySlkEmptyInputError",
    "SlkIOError",
    "HostNotReachableError",
    "SizeMismatchError",
    "check_for_errors",
]


class PySlkException(Exception):
    """A PySlkException derived from 'Exception'"""

    pass


class PySlkNoValidLoginTokenError(PySlkException):
    pass


class PySlkBadFileError(PySlkException):
    pass


class PySlkBadProcessError(PySlkException):
    pass


class ArchiveError(PySlkException):
    stdout = ["Archive failed"]
    pass


class PySlkEmptyInputError(PySlkException):
    pass


class SlkIOError(PySlkException):
    stdout = ["Unable to create namespace", "Namespace not found in list command"]
    pass


class HostNotReachableError(ArchiveError):
    stdout = ["Host not reachable"]
    pass


class SizeMismatchError(ArchiveError):
    pass


def check_for_errors(output, fun):
    """Check subprocess output for certain errors"""
    if output.returncode != 0:
        stdout = output.stdout.decode("utf-8")
        stderr = output.stderr.decode("utf-8")
        error = (
            f"pyslk.{fun}\n"
            f"args: {output.args}\n"
            f"command: {' '.join(output.args)}\n"
            f"errorcode: {output.returncode}\n"
            f"stdout: {stdout}"
            f"stderr: {stderr}"
        )

        if any([io in stdout for io in HostNotReachableError.stdout]):
            raise HostNotReachableError(error)

        if any([io in stdout for io in SlkIOError.stdout]):
            raise SlkIOError(error)

        if any([io in stdout for io in ArchiveError.stdout]):
            raise ArchiveError(error)

        # This returncode is different from the output in ~/.slk/slk-cli.log
        # we cannot distinguish archive errors by their error code yet.
        # if output.returncode == 400:
        #    raise SizeMismatchError(error)

        raise PySlkException(error)

    return output.returncode



from __future__ import annotations

import json
from typing import Union

from .file_group import FileGroup

__all__ = ["GroupCollection"]


class GroupCollection:
    def __init__(self):
        # dict has structure "index: job"
        # thus, we can have jobs without job id
        self.groups: dict[int, FileGroup] = dict()
        # iterator not initialized
        self.iterator = None

    def __iter__(self):
        self.iterator = iter(self.groups)
        return self

    def __next__(self):
        return self.iterator.__next__()

    def add_group(self, group: Union[FileGroup, dict]) -> int:
        # if group is no RecallJob but a dict, then create RecallJob from the dict
        if isinstance(group, dict):
            group = FileGroup(group)
        # add group to groups dict
        search_id: int = group.get_search_id()
        if search_id not in self.groups:
            # group does not exist yet in groups
            self.groups[search_id] = group
        else:
            # a group with the same search id exists already in groups
            self.groups[search_id] = self.groups[search_id].merge_group(group)
        # return search id of the group
        return search_id

    def get_group(self, search_id: int) -> FileGroup:
        return self.groups.get(search_id, None)

    def dump(self) -> dict[int, dict]:
        return {k: v.dump() for k, v in self.groups.items()}

    def __str__(self) -> str:
        """
        print a JSON string of all jobs in this job collection

        :return: JSON strong of all jobs in this job collection
        :rtype: str
        """
        return json.dumps({k: v.__str__() for k, v in self.groups.items()})

    def json(self) -> str:
        """
        print a JSON string of all jobs in this job collection

        :return: JSON strong of all jobs in this job collection
        :rtype: str
        """
        return self.__str__()

    def __len__(self) -> int:
        return len(self.groups)

    def size(self) -> int:
        return self.__len__()



import inspect
import json
import warnings
from pathlib import Path
from typing import Union

from ..base import StatusJob, get_resource_path
from ..constants import MAX_RETRIES_SUBMIT_JOBS_BASIC, MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE
from ..pyslk_exceptions import PySlkException
from ..raw import result_verify_job_raw, submit_verify_job_raw
from .basic import _is_job_completed, get_job_status, is_job_finished, is_job_successful
from .submitted_jobs import SubmittedJobs

__all__ = [
    "submit_verify_job",
    "get_bad_files_verify_job",
    "get_checked_resources_verify_job",
    "get_result_verify_job",
]


def submit_verify_job(
    resources: Union[str, Path, list[str], list[Path], None] = None,
    resource_ids: Union[int, str, list[int], list[str], None] = None,
    search_id: Union[int, str, None] = None,
    search_query: Union[str, None] = None,
    json_input_file: Union[str, Path, None] = None,
    recursive: bool = False,
    resume_on_page: int = 1,
    end_on_page: int = -1,
    save_mode: bool = False,
) -> SubmittedJobs:
    """
    Submit a verify job

    :param resources: provide a list of resources over which the verify job should be run
    :type resources: `str` or `Path` or `list[str]` or `list[Path]` or `None`
    :param resource_ids: provide a list of ids of resources over which the verify job should be run
    :type resource_ids: `str` or `int` or `list[str]` or `list[int]` or `None`
    :param search_id: provide a search id pointing to files over which the verify job should be run
    :type search_id: `str` or `int` or `None`
    :param search_query: provide a search query to select files over which the verify job should be run
    :type search_query: `str` or `None`
    :param json_input_file: read a JSON file which contains payload of the verify job (admin users only)
    :type json_input_file: `str` or `Path`  or `None`
    :param recursive: iterate namespaces recursively
    :type recursive: `bool`
    :param resume_on_page: resume the command and start with search result page `resume_on_page`; internally, 1000
            search results are on one 'page' and fetched by one request; you do not necessarily have read permissions
            for all of these files
    :type resume_on_page: `int`
    :param end_on_page: end with search result page `end_on_page`; internally, 1000 search results
            are on one 'page' and fetched by one request; you do not necessarily have read permissions for all of
            these files
    :type end_on_page: `int`
    :param save_mode: save mode suggested to be used in times of many timeouts; please do not regularly use this
            parameter; start one verify job per page of search results instead of one verify job for 50 pages of
            search results
    :type save_mode: `bool`
    :return: submitted jobs and, if command failed due to timeout, restart information
    :rtype: `SubmittedJobs`
    """
    first_iteration: bool = True
    iteration: int = 0
    return_code: int = -1
    submitted_jobs: SubmittedJobs = SubmittedJobs()

    # TODO: remove things

    while (first_iteration or return_code) and (
        iteration < MAX_RETRIES_SUBMIT_JOBS_BASIC + MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE
    ):
        # increment counter
        iteration = iteration + 1
        # submit verify job and get subprocess output
        if first_iteration:
            first_iteration = False
            output = submit_verify_job_raw(
                resources,
                resource_ids,
                search_id,
                search_query,
                json_input_file,
                recursive,
                resume_on_page,
                end_on_page,
                save_mode,
                json=True,
                return_type=2,
            )
        else:
            if iteration <= MAX_RETRIES_SUBMIT_JOBS_BASIC:
                warnings.warn(
                    f"pyslk.{inspect.stack()[0][3]}: 'submit_verify_job' had a timeout. Retrying ..."
                )
                output = submitted_jobs.resume_submission()
            else:
                warnings.warn(
                    f"pyslk.{inspect.stack()[0][3]}: 'submit_verify_job' had a timeout. Retrying in save mode (1000 "
                    + "files per verify job)"
                )
                output = submitted_jobs.resume_submission(save_mode=True)

        # get return code (needed for while loop)
        return_code = output.returncode

        # we have an error which is not a timeout
        if output.returncode != 0 or output.returncode != 3:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: received error code {output.returncode} on iteration {iteration}. "
                + "Error message: "
                + f"{output.stdout.decode('utf-8').rstrip()} "
                + f"{output.stderr.decode('utf-8').rstrip()}"
            )

        # generate submitted jobs object
        submitted_jobs = SubmittedJobs(
            submitted_jobs_raw_output=output.stdout.decode("utf-8").rstrip()
        )

        # warn if still timeout occurred
        if output.returncode == 3:
            warnings.warn(
                f"pyslk.{inspect.stack()[0][3]}: "
                + (MAX_RETRIES_SUBMIT_JOBS_BASIC + MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE)
                + " timeouts occurred while submitting verify jobs. Please start the remaining verify jobs at "
                + "a later point of time. Restart information is given in the Object returned by this function."
            )

        # return submitted_jobs
        return submitted_jobs


def get_result_verify_job(
    job_id: int,
    return_no_source: bool = False,
) -> dict[str, dict]:
    """
    Get results of one or more verify jobs

    :param job_id: job id
    :type job_id: `int`
    :param return_no_source: return the result of the verify job but do not contain a list of checked files
    :type return_no_source: `bool`
    :return: results of the verify job and its header
    :rtype: `dict[str, dict]`
    """

    # type check: job
    if not isinstance(job_id, int):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'job_id' has wrong type; need 'int' but got "
            + f"'{type(job_id).__name__}'"
        )

    # type check: --json-no-source
    if not isinstance(return_no_source, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'return_no_source' has wrong type; need 'bool' but got "
            + f"'{type(return_no_source).__name__}'"
        )

    return json.loads(
        result_verify_job_raw(job_id, json=True, json_no_source=return_no_source)
    )


def get_bad_files_verify_job(
    jobs: Union[int, list[int], set[int], SubmittedJobs],
) -> dict[str, list[Path]]:
    """Get results of one or more verify jobs

    :param jobs: submitted jobs (job id, list of job ids or SubmittedJobs object)
    :type jobs: `int` or `list[int]` or `set[int]` or `SubmittedJobs`
    :return: results of the verify job
    :rtype: `dict[str, list[Path]]`
    """
    # init job list
    job_ids: list[int] = []
    # tmp raw job results and paths
    tmp_results: dict[str, list[Path]] = {}
    tmp_paths: list[Path]
    # init return value
    output: dict[str, list[Path]] = {}

    # if jobs is an int, convert to list
    if isinstance(jobs, int):
        job_ids.append(jobs)
    elif isinstance(jobs, list):
        job_ids = jobs
    elif isinstance(jobs, set):
        job_ids = list(jobs)
    elif isinstance(jobs, SubmittedJobs):
        job_ids = jobs.get_job_ids()
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'jobs' has to be of type 'int', 'list[int]' or 'SubmittedJobs' but is "
            + f"{type(jobs).__name__}"
        )

    """
    Structure of output of get_result_verify_job:
    {
    "header": {
      "job_id": "187988",
      "job_status": "COMPLETED",
      "job_start": "Tue Jan 16 22:35:17 UTC 2024",
      "job_end": "Tue Jan 16 22:36:52 UTC 2024",
      "scanned_files": "10",
      "io_verify_size_failed": 2,
      "source_resources":[76311882010,76311802010,76613132010]
    },
    "erroneous_files": {
      "size_mismatch": [
        "/dkrz_test/netcdf/20240116b/file_500mb_b.nc",
        "/dkrz_test/netcdf/20240116b/file_500mb_h.nc"
      ],
      "missing": [],
      "other": []
    }
    """
    finished_job_ids: list[bool] = [is_job_finished(job_id) for job_id in job_ids]
    status_job_ids: list[StatusJob] = [get_job_status(job_id) for job_id in job_ids]
    # check if all jobs are finished; if not => exception
    if not all(finished_job_ids):
        states: list[str] = [
            f"{job_id} with status {status}"
            for job_id, finished, status in zip(
                job_ids, finished_job_ids, status_job_ids
            )
            if not finished
        ]
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: failed: some jobs are not finished: {', '.join(states)}. Please contact "
            + "support@dkrz.de if one or more stati are not QUEUED or PROCESSING."
        )
    # check if all jobs are successful or completed; if not => exception
    successful_job_ids: list[bool] = [is_job_successful(job_id) for job_id in job_ids]
    if not all(successful_job_ids):
        completed_job_ids: list[bool] = [
            _is_job_completed(job_id) for job_id in job_ids
        ]
        if not all([s or c for s, c in zip(successful_job_ids, completed_job_ids)]):
            states: list[str] = [
                f"{job_id} with status {status}"
                for job_id, successful, status in zip(
                    job_ids, successful_job_ids, status_job_ids
                )
                if not successful
            ]
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: failed: some jobs are not successful: {', '.join(states)}. Please "
                + "contact support@dkrz.de if one or more stati are not SUCCESSFUL."
            )
        # else:
        #     job_list: list[int] = [
        #         job_id
        #         for job_id, successful in zip(job_ids, successful_job_ids)
        #         if not successful
        #     ]
        #     if len(job_list) > 0:  # job_list
        #         warnings.warn(
        #             f"pyslk.{inspect.stack()[0][3]}: these jobs have the status COMPLETED but not SUCCESSFUL: "
        #             + f"{', '.join([str(job_id) for job_id in job_list])}. Please be careful when working with "
        #             + "these jobs and contact support@dkrz.de."
        #         )

    # get results of jobs
    for job_id in job_ids:
        try:
            tmp_results = get_result_verify_job(job_id, return_no_source=True)
        except PySlkException as e:
            PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: 'get_result_verify_job' failed on job {job_id}, please re-run this "
                + f"function without this job id; error details: {str(e)}"
            )

        for key in tmp_results["erroneous_files"].keys():
            tmp_paths = [
                Path(res_path) for res_path in tmp_results["erroneous_files"][key]
            ]
            if key in output:
                output[key].extend(tmp_paths)
            else:
                output[key] = tmp_paths

    # return output
    return output


def get_checked_resources_verify_job(
    jobs: Union[int, list[int], SubmittedJobs],
    force_resource_path: bool = False,
) -> list[Union[int, Path]]:
    """Get results of one or more verify jobs

    TODO: doc update

    :param jobs: submitted jobs (job id, list of job ids or SubmittedJobs object)
    :type jobs: `int` or `list[int]` or `SubmittedJobs`
    :param force_resource_path: force resource path to be returned; otherwise, resource ids will be returned
        (except if a namespace was the target of the verify job which is limited to admins)
    :type force_resource_path: `bool`
    :return: results of the verify job
    :rtype: `list[Union[int, Path]]`
    """
    # init job list
    job_ids: list[int] = []
    # tmp raw job results and paths
    tmp_resource_ids: list[int] = []
    tmp_resource_paths: list[Path] = []
    # init return value
    # TODO: int or path
    output: list[Union[int, Path]] = []

    # if jobs is an int, convert to list
    if isinstance(jobs, int):
        job_ids.append(jobs)
    elif isinstance(jobs, list):
        job_ids = jobs
    elif isinstance(jobs, SubmittedJobs):
        job_ids = jobs.get_job_ids()
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'jobs' has to be of type 'int', 'list[int]' or 'SubmittedJobs' but is "
            + f"{type(jobs).__name__}"
        )

    """
    Structure of output of get_result_verify_job:
    {
    "header": {
      "job_id": "187988",
      "job_status": "COMPLETED",
      "job_start": "Tue Jan 16 22:35:17 UTC 2024",
      "job_end": "Tue Jan 16 22:36:52 UTC 2024",
      "scanned_files": "10",
      "io_verify_size_failed": 2,
      "source_resources":[76311882010,76311802010,76613132010]
    },
    "erroneous_files": {
      "size_mismatch": [
        "/dkrz_test/netcdf/20240116b/file_500mb_b.nc",
        "/dkrz_test/netcdf/20240116b/file_500mb_h.nc"
      ],
      "missing": [],
      "other": []
    }

    OR

    {
    "header": {
      "job_id": "187988",
      "job_status": "COMPLETED",
      "job_start": "Tue Jan 16 22:35:17 UTC 2024",
      "job_end": "Tue Jan 16 22:36:52 UTC 2024",
      "scanned_files": "10",
      "io_verify_size_failed": 2,
      "source_namespace": "/arch/bb1170/b381219/DELIVERABLE_WP12-D1/output"
    },
    "erroneous_files": {
      "size_mismatch": [
        "/dkrz_test/netcdf/20240116b/file_500mb_b.nc",
        "/dkrz_test/netcdf/20240116b/file_500mb_h.nc"
      ],
      "missing": [],
      "other": []
    }
    """
    finished_job_ids: list[bool] = [is_job_finished(job_id) for job_id in job_ids]
    status_job_ids: list[StatusJob] = [get_job_status(job_id) for job_id in job_ids]
    if not all(finished_job_ids):
        states: list[str] = [
            f"{job_id} with status {status}"
            for job_id, finished, status in zip(
                job_ids, finished_job_ids, status_job_ids
            )
            if not finished
        ]
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: failed: some jobs are not finished: {', '.join(states)}. Please contact "
            + "support@dkrz.de if one or more stati are not QUEUED or PROCESSING."
        )
    # TODO: COMPLETED + SUCCESSFUL are OK; no warning when COMPLETED
    successful_job_ids: list[bool] = [is_job_successful(job_id) for job_id in job_ids]
    if not all(successful_job_ids):
        completed_job_ids: list[bool] = [
            _is_job_completed(job_id) for job_id in job_ids
        ]
        if not all([s or c for s, c in zip(successful_job_ids, completed_job_ids)]):
            states: list[str] = [
                f"{job_id} with status {status}"
                for job_id, successful, status in zip(
                    job_ids, successful_job_ids, status_job_ids
                )
                if not successful
            ]
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: failed: some jobs are not successful: {', '.join(states)}. Please "
                + "contact support@dkrz.de if one or more stati are not SUCCESSFUL."
            )
        # else:
        #     job_list: list[int] = [
        #         job_id
        #         for job_id, successful in zip(job_ids, successful_job_ids)
        #         if not successful
        #     ]
        #     if len(job_list) > 0:  # job_list
        #         warnings.warn(
        #             f"pyslk.{inspect.stack()[0][3]}: these jobs have the status COMPLETED but not SUCCESSFUL: "
        #             + f"{', '.join(job_list)}. Please be careful when working with these jobs and contact "
        #             + f"support@dkrz.de."
        #         )

    # get results of jobs
    for job_id in job_ids:
        try:
            tmp_results = get_result_verify_job(job_id, return_no_source=False)
        except PySlkException as e:
            PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: 'get_result_verify_job' failed on job {job_id}, please re-run this "
                + f"function without this job id; error details: {str(e)}"
            )
        if "source_namespace" in tmp_results["header"]:
            tmp_resource_paths.append(Path(tmp_results["header"]["source_namespace"]))
        elif "source_resources" in tmp_results["header"]:
            tmp_resource_ids.extend(tmp_results["header"]["source_resources"])

    # add namespace paths anyway
    output.extend(tmp_resource_paths)
    # if lists of resource ids were returned, we need to check whether the user accepts ids or wants paths
    if not force_resource_path:
        output.extend(tmp_resource_ids)
    else:
        output.extend(
            [get_resource_path(resource_id) for resource_id in tmp_resource_ids]
        )

    # return output
    return output



import inspect
import subprocess
from pathlib import Path
from typing import Union

from ..constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK, SLK_HELPERS
from ..pyslk_exceptions import PySlkException
from ..utils import run_slk, which

__all__ = [
    "gen_file_query_raw",
    "gen_search_query_raw",
    "searchid_exists_raw",
    "search_immediately_raw",
    "search_incomplete_raw",
    "search_raw",
    "search_status_raw",
    "search_successful_raw",
    "tnsr_raw",
    "total_number_search_results_raw",
]


def search_raw(
    search_string: Union[str, None] = None,
    group: Union[str, int, None] = None,
    user: Union[str, int, None] = None,
    name: Union[str, None] = None,
    partial: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Creates search and returns search id

    Either group, user and/or name can be set
    or a search_string can be provided.

    search_string has to be a valid JSON search string as described in the
    SLK-CLI manual. Simple double quotes have to be used in the JSON
    expression (no escaped double quotes, no escaped special characters).

    Example for search_string:
      search_string='{"resources.mtime": {"$gt": "2021-09-02"}}'

    :param search_string: JSON search query string
    :type search_string: str
    :param group: search for files belonging to the provided group name or GID
    :type group: str or int
    :param user: search for files belonging to the provided username or UID
    :type user: str or int
    :param name: search files having the provided name
    :type name: str
    :param partial: search for files with 'partial file' flag
    :type partial: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    slk_call = SLK + " search"

    if group is not None:
        if isinstance(group, (str, int)):
            slk_call = slk_call + " -group " + str(group)
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: argument "group" needs to be "str" or "int" but got '
                + f'"{type(group).__name__}".'
            )
    if user is not None:
        if isinstance(user, (str, int)):
            slk_call = slk_call + " -user " + str(user)
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: argument "user" needs to be "str" or "int" but got '
                + f'"{type(user).__name__}".'
            )
    if name is not None:
        if isinstance(name, str):
            slk_call = slk_call + " -name " + str(name)
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: argument "name" needs to be "str" but got "{type(group).__name__}".'
            )
    if partial:
        slk_call = slk_call + " -partial"
    if search_string is not None:
        if isinstance(search_string, str):
            slk_call = slk_call + f" '{search_string}'"
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: argument "search_string" needs to be "str" but got '
                + f'"{type(search_string).__name__}".'
            )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def searchid_exists_raw(
    search_id: Union[str, int], return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check if search id exists

    :param search_id: a search_id
    :type search_id: str or int
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(search_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "search_id" has to be of type "str" or "int" but is '
            + f"{type(search_id).__name__}"
        )

    slk_call = [SLK_HELPERS, "searchid_exists", str(search_id)]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def gen_file_query_raw(
    resources: Union[str, list, Path],
    recursive: bool = False,
    no_newline: bool = False,
    cached_only: bool = False,
    not_cached: bool = False,
    tape_barcodes: Union[list[str], str, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Generates a search query that searches for the listed resources

    A search query will be generated which connects all elements of
    'resources' with and 'or'. A 'resource' (an element of 'resources')
    might be one of these:

    - a filename with full path (e.g. /arch/bm0146/k204221/INDEX.txt)
    - a filename without full path (e.g. INDEX.txt)
    - a regex describing a filename (e.g. /arch/bm0146/k204221/.*.txt)
    - a namespace (e.g. /arch/bm0146/k204221 or /arch/bm0146/k204221/)

    Details are given in the slk_helpers documentation at https://docs.dkrz.de

    :param resources: list of resources to be searched for
    :type resources: str or list or Path
    :param recursive: do recursive search in the namespaces
    :type recursive: bool
    :param no_newline: do recursive search in the namespaces
    :type no_newline: bool
    :param cached_only: do recursive search in the namespaces
    :type cached_only: bool
    :param not_cached: do recursive search in the namespaces
    :type not_cached: bool
    :param tape_barcodes: do recursive search in the namespaces
    :type tape_barcodes: list[str]
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "gen_file_query"]
    if recursive:
        slk_call.append("-R")
    if isinstance(resources, str):
        slk_call.extend(resources.split(" "))
    elif isinstance(resources, Path):
        slk_call.append(str(resources))
    elif isinstance(resources, list):
        if not all([isinstance(r, (str, Path)) for r in resources]):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: if argument 'resources' is of type 'list' its items need to be"
                + "of type 'str' or Path-like but got type(s): "
                + f"'{', '.join([type(r).__name__ for r in resources if not isinstance(r, (str, Path))])}'."
            )
        slk_call.extend(resources)
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'resources' has to be 'str' or 'list' of 'str' but is "
            + f"'{type(resources).__name__}'"
        )
    if no_newline:
        slk_call.append("--no-newline")
    if cached_only:
        slk_call.append("--cached-only")
    if not_cached:
        slk_call.append("--not-cached")
    if tape_barcodes is not None:
        slk_call.append("--tape-barcodes")
        if isinstance(tape_barcodes, str):
            slk_call.extend(tape_barcodes.split(" "))
        elif isinstance(tape_barcodes, list):
            slk_call.extend(tape_barcodes)
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'tape_barcodes' has to be 'str' or 'list' of 'str' but is "
                + f"'{type(tape_barcodes).__name__}'"
            )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def gen_search_query_raw(
    key_value_pairs: Union[str, list[str]],
    recursive: bool = False,
    search_query: Union[str, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Generates a search query that searches based on the conditions provided as key-values pairs

    These key-value pairs are actually key-operator-value-triples. E.g.

    .. code-block::

        path=/arch/bm0146/k204221/iow
        resources.size < 1024
        resources.created > 2023-01-01

    Details are given in the slk_helpers documentation at https://docs.dkrz.de

    :param key_value_pairs: list of key-value pairs connected via an operator
    :type key_value_pairs: str or list
    :param recursive: do recursive search in the namespaces
    :type recursive: bool
    :param search_query: an existing search query to be extended
    :type search_query: str
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "gen_search_query"]
    if recursive:
        slk_call.append("-R")
    if isinstance(key_value_pairs, str):
        slk_call.append(key_value_pairs)
    elif isinstance(key_value_pairs, list):
        slk_call.extend(key_value_pairs)
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'key_value_pairs' has to be 'str' or 'list' of 'str' but is "
            + f"'{type(key_value_pairs).__name__}'"
        )
    if search_query is not None:
        slk_call.append("--search-query")
        if isinstance(search_query, str):
            slk_call.append(search_query)
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'search_query' has to be 'str' but is "
                + f"'{type(search_query).__name__}'"
            )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def tnsr_raw(
    search_id: Union[str, int],
    quiet: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Print the total number of search results regardless of read permissions of the current user

    :param search_id: a search_id
    :type search_id: str or int
    :param quiet: should the command be quiet or print warnings?
    :type quiet: `bool`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    return total_number_search_results_raw(search_id, quiet, return_type)


def total_number_search_results_raw(
    search_id: Union[str, int],
    quiet: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Print the total number of search results regardless of read permissions of the current user

    :param search_id: a search_id
    :type search_id: str or int
    :param quiet: should the command be quiet or print warnings?
    :type quiet: `bool`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(search_id, (str, int)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'search_id' has to be of type 'str' or 'int' but is "
            + f"{type(search_id).__name__}"
        )

    slk_call = [SLK_HELPERS, "total_number_search_results", str(search_id)]

    if quiet:
        slk_call.append("-q")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def search_immediately_raw(
    search_string: str, return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess, subprocess.Popen]:
    """
    submits a search query string to StrongLink and immediately returns the search id; search not finished

    :param search_string: JSON search query string
    :type search_string: `str`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output), 3 (running subproc.)
    :type return_type: `int`
    :returns: stdout of the slk call
    :rtype: `Union[str, int, subprocess.CompletedProcess, subprocess.Popen]`
    """
    if which(SLK_HELPERS) is None:
        raise RuntimeError(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = SLK_HELPERS + " search_immediately"

    if search_string is None:
        raise ValueError(
            f'pyslk.{inspect.stack()[0][3]}: argument "search_string" has to have a value unequal'
            ' "None"'
        )
    else:
        if isinstance(search_string, str):
            slk_call = slk_call + f" '{search_string}'"
        else:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'search_string' has wrong type; need 'str' but got "
                + f"'{type(search_string).__name__}'"
            )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    elif return_type == 3:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
            wait_until_finished=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def search_incomplete_raw(
    search_id: Union[str, int], return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check if search is incomplete

    :param search_id: a search_id
    :type search_id: `str` or `int`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: `int`
    :returns: stdout of the slk call
    :rtype: `Union[str, int, subprocess.CompletedProcess]`
    """
    if which(SLK_HELPERS) is None:
        raise RuntimeError(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(search_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "search_id" has to be of type "str" or "int" but is '
            + f"{type(search_id).__name__}"
        )
    if isinstance(search_id, str):
        try:
            int(search_id)
        except ValueError:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'search_id' cannot be processed; need 'str', which holds an "
                + f"integer number, or 'int' but got '{type(search_id).__name__}' with value '{search_id}'"
            )

    slk_call = [SLK_HELPERS, "search_incomplete", str(search_id)]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def search_status_raw(
    search_id: Union[str, int], json: bool = False, return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check if search is successful

    :param search_id: a search_id
    :type search_id: `str` or `int`
    :param json: if the output should be in json format
    :type json: `bool`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: `int`
    :returns: stdout of the slk call
    :rtype: `Union[str, int, subprocess.CompletedProcess]`
    """
    if which(SLK_HELPERS) is None:
        raise RuntimeError(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(search_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "search_id" has to be of type "str" or "int" but is '
            + f"{type(search_id).__name__}"
        )
    if isinstance(search_id, str):
        try:
            int(search_id)
        except ValueError:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'search_id' cannot be processed; need 'str', which holds an "
                + f"integer number, or 'int' but got '{type(search_id).__name__}' with value '{search_id}'"
            )

    slk_call = [SLK_HELPERS, "search_status", str(search_id)]

    # --json
    if not isinstance(json, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'json' has wrong type; need 'bool' but got "
            + f"'{type(json).__name__}'"
        )
    if json:
        slk_call.append("--json")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def search_successful_raw(
    search_id: Union[str, int], return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check if search is successful

    :param search_id: a search_id
    :type search_id: `str` or `int`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: `int`
    :returns: stdout of the slk call
    :rtype: `Union[str, int, subprocess.CompletedProcess]`
    """
    if which(SLK_HELPERS) is None:
        raise RuntimeError(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(search_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "search_id" has to be of type "str" or "int" but is '
            + f"{type(search_id).__name__}"
        )
    if isinstance(search_id, str):
        try:
            int(search_id)
        except ValueError:
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: 'search_id' cannot be processed; need 'str', which holds an "
                + f"integer number, or 'int' but got '{type(search_id).__name__}' with value '{search_id}'"
            )

    slk_call = [SLK_HELPERS, "search_successful", str(search_id)]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



import inspect

from .config import get as get_config
from .constants import SLK, SLK_HELPERS
from .pyslk_exceptions import PySlkException
from .utils import run_slk, which

__all__ = [
    "slk_helpers_version",
    "slk_version",
]

__slk_version__ = get_config("__slk_version__")
__slk_helpers_version__ = get_config("__slk_helpers_version__")

# this is the slk_version that we used to develop parsing of command line output
# if the command line output format changes in the future, the parser has to
# be adapted.
__slk_version_used_for_parser_syntax__ = "3.3.91"
__slk_helpers_version_used_for_parser_syntax__ = "1.13.3"


def slk_version() -> str:
    """List the version of slk

    :returns: stdout of the slk call
    :rtype: str
    """

    if which(SLK) is None:
        raise PySlkException("pyslk: " + SLK + ": command not found")

    slk_call = [SLK, "version"]

    return run_slk(slk_call, inspect.stack()[0][3])


def slk_helpers_version() -> str:
    """List the version of slk_helpers

    :returns: stdout of the slk_helpers call
    :rtype: str
    """

    if which(SLK_HELPERS) is None:
        raise PySlkException("pyslk: " + SLK_HELPERS + ": command not found")

    slk_call = [SLK_HELPERS, "version"]

    return run_slk(slk_call, inspect.stack()[0][3])



import glob
from os import path as op

from ..raw import archive_raw

__all__ = [
    "archive",
]


def archive(
    resource, dst_gns, recursive=False, retry=False, **kwargs
) -> tuple[list, str]:
    """Upload data to tape archive using slk archive

    This archiving function wraps some functionalities around :py:meth:`~pyslk.archive_raw`.
    It returns a list of archived files or directories and can use
    `tenacity <https://github.com/jd/tenacity>`_ for retrying failed archive attempts.

    :param resource:
        Path, pattern or list of files that should be archived.
    :type resource: ``str`` or ``list``
    :param dst_gns:
        Destination directory for archiving.
    :type dst_gns: ``str``
    :param recursive:
        Archive recursively.
    :type recursive: ``bool``
    :param retry:
        Retry archiving if an ``ArchiveError`` is encountered.
        If ``retry`` is True and archiving fails due to ``ArchiveError``,
        archiving will wait for 10 seconds and retry. Maximal 3 attempts are made.
        After that, ``ArchiveError`` will be raised after all.
        Requires ``tenacity`` to be installed.
    :type retry: ``bool``
    :returns:
        a tuple containing archived files and stdout from archive_raw.
    :rtype: ``(list, str)``

    .. seealso::
        * :py:meth:`~pyslk.archive_raw`

    """

    if not isinstance(resource, type([])):
        resource = [resource]

    resources = []
    for r in resource:
        resources.extend(map(op.abspath, glob.glob(r)))

    # remove possible duplicates
    resources = list(dict.fromkeys(resources))
    # create paths of archived files for return
    arch_file_paths = [op.join(dst_gns, op.basename(r)) for r in resources]

    if retry is True:
        from ..retry import try_archive

        return arch_file_paths, try_archive(
            resources, dst_gns, recursive=recursive, **kwargs
        )
    return arch_file_paths, archive_raw(
        resources, dst_gns, recursive=recursive, **kwargs
    )



import inspect
import math
import os
import subprocess
import typing
from pathlib import Path
from typing import Union

from ..base import get_resource_path, ls
from ..constants import PYSLK_FILE_SIZE_UNITS
from ..pyslk_exceptions import PySlkException
from ..raw import (
    checksum_raw,
    chmod_raw,
    delete_raw,
    exists_raw,
    group_raw,
    has_no_flag_partial_raw,
    list_raw,
    mkdir_raw,
    move_raw,
    owner_raw,
    rename_raw,
    resource_mtime_raw,
    resource_permissions_raw,
    resource_tape_raw,
    size_raw,
)
from ..utils import _convert_size, _parse_list_to_rows, _parse_size
from .resource_extras import (
    _check_resource_and_search_id,
    is_namespace,
    resource_exists,
)

__all__ = [
    "access_hsm",
    "arch_size",
    "chgrp",
    "chmod",
    "chown",
    "delete",
    "get_checksum",
    "get_resource_id",
    "get_resource_mtime",
    "get_resource_permissions",
    "get_resource_size",
    "get_resource_tape",
    "has_no_flag_partial",
    "has_no_flag_partial_details",
    "makedirs",
    "mkdir",
    "move",
    "rename",
]


def get_resource_tape(
    resource_path: Union[str, Path],
) -> typing.Optional[dict[int, str]]:
    """returns tape on which resource with given path is stored on

    :param resource_path: namespace or resource
    :type resource_path: ``str`` or ``path-like``
    :returns: tape(s) on which a file is/are stored on as dict; None otherwise
    :rtype: ``dict[int, str]`` or ``None``
    """
    if not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource_path' has wrong type; need 'str' or 'Path' but "
            + f"got {type(resource_path).__name__}"
        )
    output: subprocess.CompletedProcess = resource_tape_raw(
        str(resource_path), return_type=2
    )

    if output.returncode == 0:
        tmp_output: list[str] = [
            i for i in output.stdout.decode("utf-8").rstrip().split("\n") if i != ""
        ]
        try:
            n_tapes = int(tmp_output[-1].split(": ")[1])
        except ValueError:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: unexpected output from 'slk_helpers resource_tape': "
                + f"{output.stdout.decode('utf-8').rstrip()} "
                + f"{output.stderr.decode('utf-8').rstrip()}"
            )
        # file stored on no tape
        if n_tapes == 0:
            return None
        if len(tmp_output) != n_tapes + 1:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: unexpected output from 'slk_helpers resource_tape': "
                + f"{output.stdout.decode('utf-8').rstrip()} "
                + f"{output.stderr.decode('utf-8').rstrip()}"
            )
        out_dict: dict = {}
        for t in tmp_output[:-1]:
            t = t.split(": ")[1].split("(")
            out_dict[int(t[0])] = t[1][:-1]
        return out_dict
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def get_resource_id(resource_path: Union[str, Path]) -> typing.Optional[int]:
    """returns resource_id to a resource path

    :param resource_path: namespace or resource
    :type resource_path: ``str`` or ``path-like``
    :returns: resource_id if the file exists; None otherwise
    :rtype: ``int`` or ``None``
    """
    if not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource_path' has wrong type; need 'str' or 'Path' but "
            + f"got {type(resource_path).__name__}"
        )
    output: subprocess.CompletedProcess = exists_raw(str(resource_path), return_type=2)
    if output.returncode == 0:
        return int(output.stdout.decode("utf-8").rstrip().split(":")[1])
    elif output.returncode == 1:
        return None
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def get_resource_mtime(resource_path: Union[str, Path]) -> typing.Optional[int]:
    """returns mtime of a resource path

    :param resource_path: namespace or resource
    :type resource_path: ``str`` or ``path-like``
    :returns: mtime if the file exists; None otherwise
    :rtype: ``int`` or ``None``
    """
    if not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource_path' has wrong type; need 'str' or 'Path' but "
            + f"got {type(resource_path).__name__}"
        )
    output: subprocess.CompletedProcess = resource_mtime_raw(
        str(resource_path), return_type=2
    )
    if output.returncode == 0:
        return int(output.stdout.decode("utf-8").rstrip())
    elif output.returncode == 1:
        return None
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def get_resource_size(
    resource: Union[str, int, Path], recursive: bool = False
) -> typing.Optional[int]:
    """Returns file size in byte

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param recursive: use the -R to calculate size recursively
    :type recursive: `bool`
    :returns: size in byte; None if resource does not exist
    :rtype: ``int`` or ``None``
    """
    if not resource_exists(resource):
        return None
    resource_id: Union[int, None] = None
    resource_path: Union[Path, None] = None
    if isinstance(resource, int):
        resource_id = resource
    elif isinstance(resource, Path):
        resource_path = resource
    elif isinstance(resource, str):
        # resource can either be a resource id or a resource path
        try:
            resource_id = int(resource)
        except ValueError:
            try:
                resource_path = Path(resource)
            except ValueError:
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: cannot convert 'resource' to 'int' (== resource id) "
                    + "or 'path-like' (== resource path)"
                )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' has wrong type; need 'str', 'int' or 'Path' but "
            + f"got {type(resource).__name__}"
        )
    if resource_path is not None and not resource_exists(resource_path):
        raise FileNotFoundError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' was considered as resource path and does not exist: {resource}"
        )
    if resource_id is not None and not get_resource_path(resource_id):
        raise FileNotFoundError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' was considered as resource id and does not exist: {resource}"
        )
    # check type of 'recursive'
    if not isinstance(recursive, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'recursive' needs to be 'bool' but got "
            + f"'{type(recursive).__name__}'"
        )
    return int(size_raw(resource_path, resource_id, recursive))


def get_resource_permissions(
    resource: Union[str, int, Path, None] = None,
    as_octal_number: bool = False,
) -> typing.Optional[Union[str, bool]]:
    """Get path for a resource id

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param as_octal_number: Do not return the permissions as combination of x/w/r/- but as three digit octal number
    :type as_octal_number: ``bool``
    :returns: permissions string; False if resource does not exist
    :rtype: ``str`` or ``bool`` or ``None``
    """
    if not resource_exists(resource):
        return None
    resource_id: Union[int, None] = None
    resource_path: Union[Path, None] = None
    if isinstance(resource, int):
        resource_id = resource
    elif isinstance(resource, Path):
        resource_path = resource
    elif isinstance(resource, str):
        # resource can either be a resource id or a resource path
        try:
            resource_id = int(resource)
        except ValueError:
            try:
                resource_path = Path(resource)
            except ValueError:
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: cannot convert 'resource' to 'int' (== resource id) "
                    + "or 'path-like' (== resource path)"
                )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource' has wrong type; need 'str', 'int' or 'Path' but "
            + f"got {type(resource).__name__}"
        )
    output: subprocess.CompletedProcess = resource_permissions_raw(
        resource_path, resource_id, as_octal_number, return_type=2
    )
    if output.returncode == 0:
        return output.stdout.decode("utf-8").rstrip()
    elif output.returncode == 1:
        return None
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def access_hsm(
    resource: Union[list[str], list[Path], str, Path], mode: int
) -> Union[bool, list[bool]]:
    if isinstance(resource, str):
        return access_hsm(Path(resource), mode)
    elif isinstance(resource, list):
        return [access_hsm(i, mode) for i in resource]
    elif isinstance(resource, Path):
        if mode not in [os.F_OK, os.R_OK, os.W_OK, os.X_OK]:
            return [False]
        if not resource_exists(resource):
            return [False]
        if mode == os.F_OK:
            return [True]
        # TODO: can we use resource_permissions here?
        raise NotImplementedError(
            "Tests for 'os.W_OK', 'os.X_OK' and 'os.R_OK' are not implemented yet."
        )
    else:
        raise TypeError(
            "Need 'str', path-like object or list of one of both types as input. Got "
            + f"'{type(resource).__name__}'."
        )


def arch_size(
    path_or_id: Union[str, int], unit: str = "B"
) -> dict[str, Union[str, float, int]]:
    """Get archive size from search id or GNS path
    by recursively listing all files of archive and
    adding file sizes

    :param path_or_id: search id or gns path
    :type path_or_id: ``str``
    :param unit: Prefix of returned size must be
        one of B, K, M, G, T, P or h for Byte, Kilobyte, Megabyte
        Gigabyte, Terrabyte, Petabyte or "human-readable";
        default: B
    :type unit: ``str``
    :returns: archive size, in key "value" contains size without unit and key "unit" contains unit
    :rtype: ``dict``
    """
    if not isinstance(unit, str):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_format' has to be a string of value "
            + f"{', '.join(PYSLK_FILE_SIZE_UNITS.keys())} or h."
        )

    # Get archive files
    list_out = list_raw(path_or_id, recursive=True)
    # df = pd.read_fwf(list_out, col_specs = 'infer',
    #                 width=column_widths, dtype=str, header=None,
    #                 names=column_names, infer_nrows=10000) #infer_nrows
    # df.drop(df.tail(2).index, inplace=True)
    rows = _parse_list_to_rows(
        list_out.split("\n"), path_or_id=path_or_id, full_path=False
    )
    sizes = [_parse_size(r[3]) for r in rows]

    # sum sizes
    size_byte = sum([s for s in sizes if not math.isnan(s)])

    # check if return format is byte or 'best size'/human-readable
    return _convert_size(size_byte, unit)


def mkdir(gns_path: Union[str, Path]) -> int:
    """Create a directory

    If the directory already exists, ``FileExistsError`` is raised. If a parent directory in the path does not exist,
    ``FileNotFoundError`` is raised.

    :param gns_path: gns path to create
    :type gns_path: ``str`` or ``Path``
    :returns: namespace/resource id of the created namespace
    :rtype: ``int``

    .. seealso::
        * :py:meth:`~pyslk.makedirs`
        * :py:meth:`~pyslk.mkdir_raw`
    """
    # throw FileExistsError if resource already exists
    if resource_exists(gns_path):
        raise FileExistsError(f"pyslk.{inspect.stack()[0][3]}: file exists: {gns_path}")
    # check if parent directory does not exist
    if not resource_exists(os.path.dirname(gns_path)):
        raise FileNotFoundError(
            f"pyslk.{inspect.stack()[0][3]}: no such file or directory: {os.path.dirname(gns_path)}"
        )
    output: subprocess.CompletedProcess = mkdir_raw(
        gns_path, recursive=False, parents=False, return_type=2
    )
    # if problems during creation of folder / namespace
    if output.returncode > 0:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )
    return get_resource_id(gns_path)


def makedirs(gns_path: Union[str, Path], exist_ok: bool = False) -> int:
    """Create a directory like 'mkdir()' but create parent directories recursively, if they do not exist

    If exist_ok is False (the default), a FileExistsError is raised if the target directory already exists.

    :param gns_path: gns path to create
    :type gns_path: ``str`` or ``Path``
    :param exist_ok: throw no error if folder already exists (like 'mkdir -p')
    :type exist_ok: ``bool``
    :returns: namespace/resource id of the created namespace
    :rtype: ``int``

    .. seealso::
        * :py:meth:`~pyslk.mkdir`
        * :py:meth:`~pyslk.mkdir_raw`
    """
    # throw FileExistsError if resource already exists and [it is not a namespace or exist_ok is not set]
    if resource_exists(gns_path) and (not is_namespace(gns_path) or not exist_ok):
        raise FileExistsError(f"pyslk.{inspect.stack()[0][3]}: file exists: {gns_path}")
    # following if-clause is actually not needed because we catch the different cases already in the lines above
    if exist_ok:
        # create recursively and do nothing if namespace exists
        output: subprocess.CompletedProcess = mkdir_raw(
            gns_path, recursive=False, parents=True, return_type=2
        )
    else:
        # create recursively and throw error if namespace exists
        output: subprocess.CompletedProcess = mkdir_raw(
            gns_path, recursive=True, parents=False, return_type=2
        )
    # if problems during creation of folder / namespace
    if output.returncode > 0:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )
    return get_resource_id(gns_path)


def rename(old_name: str, new_name: str) -> int:
    """Rename a folder or file; moving is not possible

    :param old_name: folder or file name (full GNS path)
    :type old_name: ``str``
    :param new_name: new name (only name; no full GNS path)
    :type new_name: ``str``
    :returns: return resource id of the renamed resource
    :rtype: ``int``
    """
    # get res id and throw error if source file does not exist
    resource_id: Union[int, None] = get_resource_id(old_name)
    if resource_id is None:
        raise FileNotFoundError(
            f"pyslk.{inspect.stack()[0][3]}: resource does not exist: {old_name}"
        )
    # run command
    output: subprocess.CompletedProcess = rename_raw(old_name, new_name, return_type=2)
    # if problems during creation of folder / namespace
    if output.returncode > 0:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )
    # resource id
    return resource_id


def move(src_path: str, dst_gns: str, no_overwrite: bool) -> int:
    """Move namespaces/files from one parent folder to another; renaming is
    not possible

    :param src_path: namespace or file (full GNS path)
    :type src_path: ``str``
    :param dst_gns: new parent namespace
    :type dst_gns: ``str``
    :param no_overwrite: do not overwrite target file if it exists
    :type no_overwrite: ``bool``
    :returns: return resource id of the moved resource
    :rtype: ``int``
    """
    # get res id and throw error if source file does not exist
    resource_id: Union[int, None] = get_resource_id(src_path)
    if resource_id is None:
        raise FileNotFoundError(
            f"pyslk.{inspect.stack()[0][3]}: resource does not exist: {src_path}"
        )
    # run command
    output: subprocess.CompletedProcess = move_raw(
        src_path, dst_gns, no_overwrite, return_type=2
    )
    # if problems during creation of folder / namespace
    if output.returncode > 0:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )
    # return resource id
    return resource_id


def delete(gns_path: Union[str, Path, list], recursive: bool = False) -> None:
    """Soft delete a namespace (optionally all child objects of a non-empty
            namespace) or a specific file

    :param gns_path: namespace or file (full GNS path); can be file list
    :type gns_path: ``str`` or list or ``Path``
    :param recursive: use the -R flag to delete recursively, Default: False
    :type recursive: ``bool``
    :returns: nothing is returned (void function)
    :rtype: ``None``
    """
    # TODO: What to do when some files of LIST exist but not all?
    # throw error if source file does not exist
    if isinstance(gns_path, (str, Path)) and not resource_exists(gns_path):
        raise FileNotFoundError(
            f"pyslk.{inspect.stack()[0][3]}: resource does not exist: {gns_path}"
        )
    # check type of 'recursive'
    if not isinstance(recursive, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'recursive' needs to be 'bool' but got "
            + f"'{type(recursive).__name__}'"
        )
    # run command
    delete_raw(gns_path, recursive)


def get_checksum(
    resource: Union[str, Path],
    checksum_type: Union[str, None] = None,
) -> typing.Optional[dict[str, str]]:
    """Get a checksum of a resource

    :param resource: resource (full path)
    :type resource: ``str`` or ``Path``
    :param checksum_type: checksum_type (possible values: None, "sha512", "adler32"; None => print all)
    :type checksum_type: ``str``
    :returns: dictionary with checksum type as key(s) and checksum(s) as value(s); empty keys if no checksum; 'None' if
        resource does not exist
    :rtype: ``dict`` or ``None``
    """
    # throw error if source file does not exist
    if not resource_exists(resource):
        return None
    # run command
    output: subprocess.CompletedProcess = checksum_raw(
        resource, checksum_type, return_type=2
    )
    if output.returncode == 0:
        if checksum_type is None:
            return {
                o.split(": ")[0]: o.split(": ")[1]
                for o in output.stdout.decode("utf-8").split("\n")
                if o != ""
            }
        else:
            return {checksum_type: output.stdout.decode("utf-8").rstrip()}
    elif output.returncode == 1:
        return None
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def _has_no_flag_partial(
    resource: Union[str, int, Path, list, None] = None,
    search_id: Union[str, int, None] = None,
    details: bool = False,
) -> typing.Optional[Union[bool, dict[str, list[Path]]]]:
    """Check if whether file(s) is/are flagged as partial; return True/False or a dict with keys 'flag_partial' and
        'no_flag_partial', depending on 'details'

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :param details: print flagging status of each file; if not set, output is False if no file in flagged and else True
    :type details: ``bool``
    :returns: True if one or more files are flagged; False otherwise; if 'details': dictionary with two keys
        'flag_partial' and 'no_flag_partial' which each have a list of files as value
    :rtype: ``bool`` or ``dict`` or ``None``
    """
    # check a parse the input values
    check_result = _check_resource_and_search_id(
        inspect.stack()[0][3], resource, search_id
    )
    # if we received None=> return None
    if check_result is None:
        return None
    # extract values from resource path, resource id and search id
    resource_path, resource_id, search_id = check_result
    # run raw function
    output: subprocess.CompletedProcess = has_no_flag_partial_raw(
        resource_path,
        resource_id,
        search_id,
        recursive=True,
        verbose=False,
        double_verbose=details,
        return_type=2,
    )
    if output.returncode in [0, 1]:
        if not details:
            if output.returncode == 0:
                return True
            elif output.returncode == 1:
                return False
        else:
            # [:-1] because we drop the last row
            return {
                "no_flag_partial": [
                    Path(o.split(" ")[0])
                    for o in output.stdout.decode("utf-8").split("\n")[:-2]
                    if len(o) > 0 and "has no partial flag" in o
                ],
                "flag_partial": [
                    Path(o.split(" ")[0])
                    for o in output.stdout.decode("utf-8").split("\n")[:-2]
                    if len(o) > 0 and "has partial flag" in o
                ],
            }
    else:
        # output.returncode is neither 0 nor 1
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def has_no_flag_partial(
    resource: Union[str, int, Path, list, None] = None,
    search_id: Union[str, int, None] = None,
) -> bool:
    """Check if whether file(s) is/are flagged as partial; return True/False

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :returns: True if no files is flagged; False otherwise
    :rtype: ``bool``
    """
    return _has_no_flag_partial(resource, search_id, details=False)


def has_no_flag_partial_details(
    resource: Union[str, int, Path, list, None] = None,
    search_id: Union[str, int, None] = None,
) -> dict[str, list[Path]]:
    """Check if whether file(s) is/are flagged as partial; returns dict with keys 'flag_partial' and 'no_flag_partial'

    :param resource: a resource id or a resource path
    :type resource: ``str`` or ``int`` or ``path-like``
    :param search_id: id of a search
    :type search_id: ``int`` or ``str``
    :returns: dictionary with two keys 'flag_partial' and 'no_flag_partial' which each have a list of files as value
    :rtype: ``dict[str, list[Path]]``
    """
    return _has_no_flag_partial(resource, search_id, details=True)


def chown(
    gns_path: [str, Path], owner: Union[str, int], recursive: bool = False
) -> typing.Optional[dict]:
    """Change the owner of a resource or namespace

    :param gns_path: namespace or file (full GNS path)
    :type gns_path: ``str`` or ``Path``
    :param owner: new owner of a file (username or uid)
    :type owner: ``str`` or ``int``
    :param recursive: use the -R flag to delete recursively, Default: False
    :type recursive: ``bool``
    :returns: dict of Paths of the modified files 'PATH: TRUE-IF-OWNER-CORRECT'; None if gns_path does not exist
    :rtype: ``dict`` or ``None``
    """
    # check input types
    if not isinstance(owner, (str, int)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'owner' needs to be 'str' or 'int' but got "
            + f"'{type(owner).__name__}'"
        )
    try:
        owner = int(owner)
    except ValueError:
        pass
    if isinstance(gns_path, (str, Path)):
        # if path_or_id is a str or Path check whether the resource exists
        if not resource_exists(gns_path):
            return None
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str' or Path-like but got "
            + f"'{type(gns_path).__name__}'"
        )
    # check type of 'recursive'
    if not isinstance(recursive, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'recursive' needs to be 'bool' but got "
            + f"'{type(recursive).__name__}'"
        )
    # run tag_raw
    output_raw: subprocess.CompletedProcess = owner_raw(
        gns_path, owner, recursive, return_type=2
    )
    # create actual output
    output: dict[str, Union[bool, str, list]] = dict()
    # copy exit code and stdout/stderr
    output["error"] = output_raw.returncode != 0
    output["stdout"] = output_raw.stdout.decode("utf-8").rstrip()
    output["stderr"] = output_raw.stderr.decode("utf-8").rstrip()
    # collect modified / not-modified files
    if isinstance(owner, int):
        files = ls(gns_path, recursive=recursive, numeric_ids=True)
    else:
        files = ls(gns_path, recursive=recursive, numeric_ids=False)
    output["files_correct"] = list(files.loc[files["owner"] == owner, "filename"])
    output["files_incorrect"] = list(files.loc[files["owner"] != owner, "filename"])
    # return output
    return output


def chgrp(
    gns_path: str, group: Union[str, int], recursive: bool = False
) -> typing.Optional[dict]:
    """Change the group of a resource or namespace

    :param gns_path: namespace or file (full GNS path)
    :type gns_path: ``str``
    :param group: new group of a file (group name or gid)
    :type group: ``str`` or ``int``
    :param recursive: use the -R flag to delete recursively, Default: False
    :type recursive: ``bool``
    :returns: dict with stdout/stderr, exit code and lists of files with correct and incorrect group;
                None if gns_path does not exist
    :rtype: ``dict`` or ``None``
    """
    # check input types
    if not isinstance(group, (str, int)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'group' needs to be 'str' or 'int' but got "
            + f"'{type(group).__name__}'"
        )
    try:
        group = int(group)
    except ValueError:
        pass
    if isinstance(gns_path, (str, Path)):
        # if path_or_id is a str or Path check whether the resource exists
        if not resource_exists(gns_path):
            return None
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str' or Path-like but got "
            + f"'{type(gns_path).__name__}'"
        )
    # check type of 'recursive'
    if not isinstance(recursive, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'recursive' needs to be 'bool' but got "
            + f"'{type(recursive).__name__}'"
        )
    # run tag_raw
    output_raw: subprocess.CompletedProcess = group_raw(
        gns_path, group, recursive, return_type=2
    )
    # create actual output
    output: dict[str, Union[bool, str, list]] = dict()
    # copy exit code and stdout/stderr
    output["error"] = output_raw.returncode != 0
    output["stdout"] = output_raw.stdout.decode("utf-8").rstrip()
    output["stderr"] = output_raw.stderr.decode("utf-8").rstrip()
    # collect modified / not-modified files
    if isinstance(group, int):
        files = ls(gns_path, recursive=recursive, numeric_ids=True)
    else:
        files = ls(gns_path, recursive=recursive, numeric_ids=False)
    output["files_correct"] = list(files.loc[files["group"] == group, "filename"])
    output["files_incorrect"] = list(files.loc[files["group"] != group, "filename"])
    # return output
    return output


def chmod(
    gns_path: Union[str, list], mode: Union[str, int], recursive: bool = False
) -> typing.Optional[bool]:
    """Change the access mode of a resource or namespace

    :param gns_path: namespace or file (full GNS path); can be file list
    :type gns_path: ``str`` or ``list``
    :param mode: new mode/permissions of a file (as known from bash's chmod)
    :type mode: ``str`` or ``int``
    :param recursive: use the -R flag to delete recursively, Default: False
    :type recursive: ``bool``
    :returns: True if successful, None if target does not exist, PySlkException if fails
    :rtype: ``bool`` or ``None``
    """
    # check input types
    if not isinstance(mode, (str, int)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'mode' needs to be 'str' or 'int' but got "
            + f"'{type(mode).__name__}'"
        )
    try:
        mode = int(mode)
    except ValueError:
        pass
    if isinstance(gns_path, (str, Path)):
        # if path_or_id is a str or Path check whether the resource exists
        if not resource_exists(gns_path):
            return None
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str' or Path-like but got "
            + f"'{type(gns_path).__name__}'"
        )
    # check type of 'recursive'
    if not isinstance(recursive, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'recursive' needs to be 'bool' but got "
            + f"'{type(recursive).__name__}'"
        )
    # run tag_raw
    output: subprocess.CompletedProcess = chmod_raw(
        gns_path, mode, recursive, return_type=2
    )
    # check if error occurred
    if output.returncode == 0:
        return True
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )



from .basic import (
    _is_job_completed,
    get_job_status,
    is_job_finished,
    is_job_processing,
    is_job_queued,
    is_job_successful,
    job_exists,
    job_queue,
)
from .submitted_jobs import SubmittedJobs
from .verify import (
    get_bad_files_verify_job,
    get_checked_resources_verify_job,
    get_result_verify_job,
    submit_verify_job,
)

__all__ = [
    "SubmittedJobs",
    "_is_job_completed",
    "get_bad_files_verify_job",
    "get_checked_resources_verify_job",
    "get_job_status",
    "get_result_verify_job",
    "is_job_finished",
    "is_job_processing",
    "is_job_queued",
    "is_job_successful",
    "job_exists",
    "job_queue",
    "submit_verify_job",
]



from warnings import warn

from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed

from .pyslk_exceptions import ArchiveError
from .raw import archive_raw

__all__ = ["try_archive", "log_attempt_number"]

archive_attempts = 3


def log_attempt_number(retry_state):
    """return the result of the last call attempt"""
    warn(f"{retry_state.outcome}, retrying: {retry_state.attempt_number}...")


@retry(
    stop=stop_after_attempt(archive_attempts),
    retry=retry_if_exception_type(ArchiveError),
    reraise=True,
    wait=wait_fixed(10),
    after=log_attempt_number,
)
def try_archive(resource, dst_gns, recursive=False, **kwargs):
    """Archive function wrapper for slk_archive

    If the archive fails due to HostNotReachableError, archiving will
    wait for 10 seconds and retry. Maximal 3 attempts are made.
    After that, HostNotReachableError will be raised after all.

    """
    return archive_raw(resource, dst_gns, recursive=recursive, **kwargs)



import inspect
import subprocess
import typing
from pathlib import Path
from typing import Union

from ..pyslk_exceptions import PySlkException
from ..raw import resource_path_raw

__all__ = ["get_resource_path"]


def get_resource_path(resource_id: Union[str, int]) -> typing.Optional[Path]:
    """Get path for a resource id

    :param resource_id: a resource_id
    :type resource_id: ``str`` or ``int``
    :returns: path of the resource; None if resource does not exist
    :rtype: ``Path`` or ``None``
    """
    if not isinstance(resource_id, (str, int)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource_id' has wrong type; need 'str' or 'int' but "
            + f"got {type(resource_id).__name__}"
        )
    try:
        resource_id = int(resource_id)
    except ValueError:
        return None
    output: subprocess.CompletedProcess = resource_path_raw(resource_id, return_type=2)
    if output.returncode == 0:
        return Path(output.stdout.decode("utf-8").rstrip())
    elif output.returncode == 1:
        return None
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )



import math
import os

import pytest

from pyslk.utils import _parse_list_to_rows, _parse_size

test_dir = os.path.dirname((os.path.abspath(__file__)))


@pytest.fixture
def test_output():
    filename = os.path.join(test_dir, "test_dir_output.txt")
    with open(filename) as f:
        test_str = f.read()
    return test_str


def test_parse_list_to_rows(test_output):
    rows = _parse_list_to_rows(test_output.split("\n"))
    assert rows[273][8] == "uu0808"


def test_parse_size():
    assert _parse_size("22B") == 22.0
    assert _parse_size("22M") == 22 * 1.0e6
    assert _parse_size("22.3M") == 22.3 * 1.0e6
    assert _parse_size("22.3G") == 22.3 * 1.0e9
    assert math.isnan(_parse_size("22.3X"))
    assert math.isnan(_parse_size(""))



import inspect
import subprocess
import warnings
from pathlib import Path
from typing import Union

import pandas as pd

from ..constants import PYSLK_DEFAULT_LIST_CLONE_COLUMNS, PYSLK_DEFAULT_LIST_COLUMNS
from ..pyslk_exceptions import PySlkException
from ..raw import list_clone_file_raw, list_clone_search_raw, list_raw
from ..utils import _parse_list

__all__ = ["ls", "list_clone_search", "list_clone_file"]


def ls(
    path_or_id: Union[
        str, int, Path, list[str], list[int], list[Path], set[str], set[int], set[Path]
    ],
    show_hidden: bool = False,
    numeric_ids: bool = False,
    recursive: bool = False,
    column_names: list = PYSLK_DEFAULT_LIST_COLUMNS,
    parse_dates: bool = True,
    parse_sizes: bool = True,
    full_path: bool = True,
) -> pd.DataFrame:
    """Return pandas.DataFrame containing results from search id or GNS path

    Calls :py:meth:`~pyslk.list_raw` and parses the return string into a
    ``pandas.DataFrame``. All arguments are copied 1:1 except for 'column_widths'.
    Assumes six output columns of 'slk list' having
    the widths 12, 12, 12, 9, 4, 4, 5 and 999 (999 => as wide as necessary).

    Note: :py:meth:`~pyslk.list_raw` currently only print the modification date and no
    modification time. The output of this parser might be modified in
    future when modification times are printed as well.

    :param path_or_id: search id or gns path
    :type path_or_id: ``str`` or ``Path`` or ``int``
    :param show_hidden: show '.' files, default: False (don't show these files)
    :type show_hidden: ``bool``
    :param numeric_ids: show numeric values for user and group, default: False
        (show user and group names)
    :type numeric_ids: ``bool``
    :param recursive: use the -R flag to list recursively, default: False
    :type recursive: ``bool``
    :param column_names: names of the columns in the pandas.DataFrame
    :type column_names: ``list``
    :param parse_dates: parse day, month and year into a datetime column.
    :type parse_dates: ``bool``
    :param parse_sizes: parse 'filesize' column into bytes integer.
    :type parse_sizes: ``bool``
    :param full_path: add full filepath to filename column.
    :type full_path: ``bool``
    :returns: output of 'slk list' parsed into a pandas.DataFrame with eight
        columns (permissions, owner, group, size, day, month, year, filename)
    :rtype: ``pandas.DataFrame``

    .. seealso::
        * :py:meth:`~pyslk.list_raw`

    """
    # try:
    #     import pandas as pd
    # except ModuleNotFoundError:
    #     PySlkException(f"pyslk.{inspect.stack()[0][3]}: functions needs "
    #                    "'pandas' to be installed")

    # list_out = io.StringIO(list_raw(path_or_id, show_hidden=show_hidden, numeric_ids=numeric_ids, R=R, text=text)+"\n"
    #                       "-----------    user      group             0   28"
    #                       " Aug 1986  file")
    if isinstance(path_or_id, (list, set)):
        return pd.concat(
            [
                ls(
                    p,
                    show_hidden=show_hidden,
                    numeric_ids=numeric_ids,
                    recursive=recursive,
                    column_names=column_names,
                    parse_dates=parse_dates,
                )
                for p in path_or_id
            ],
            ignore_index=True,
        )

    output: subprocess.CompletedProcess = list_raw(
        path_or_id,
        show_hidden=show_hidden,
        numeric_ids=numeric_ids,
        recursive=recursive,
        text=False,
        print_bytes=parse_sizes,
        return_type=2,
    )
    stdout = output.stdout.decode("utf-8").rstrip()
    stderr = output.stderr.decode("utf-8").rstrip()
    if output.returncode > 0:
        if "ERROR: No resources found for given search id:" in f"{stdout}  {stderr}":
            return pd.DataFrame(columns=column_names)
        else:
            raise PySlkException(f"pyslk.{inspect.stack()[0][3]}: {stdout} {stderr}")
    # df = pd.read_fwf(list_out, col_specs = 'infer',
    #                 width=column_widths, dtype=str, header=None,
    #                 names=column_names, infer_nrows=10000) #infer_nrows
    # df.drop(df.tail(2).index, inplace=True)

    df = _parse_list(
        stdout,
        path_or_id,
        column_names,
        parse_dates,
        parse_sizes,
        numeric_ids,
        full_path,
    )

    return df


def list_clone_search(
    search_id: Union[str, int, list[int], list[str], set[int], set[str]],
    only_files: bool = False,
    only_namespaces: bool = False,
    start: Union[int, None] = None,
    count: Union[int, None] = None,
    column_names=PYSLK_DEFAULT_LIST_CLONE_COLUMNS,
) -> pd.DataFrame:
    """Return pandas.DataFrame containing results from search id

    Calls :py:meth:`~pyslk.list_clone_search_raw` and parses the return string into
    a :py:meth:`pandas.DataFrame`. Assumes six output columns of 'slk_helpers list_clone_search'
    having the widths 12, 16, 999 (999 => as wide as necessary).

    :param search_id: search id of search which results should be printed
    :type search_id: ``str``, ``int``, ``list``, ``set``
    :param only_files: print only files (like default for `slk list`)
    :type only_files: ``bool``
    :param only_namespaces: print only namespaces
    :type only_namespaces: ``bool``
    :param start: collect search results starting with the result 'start'; if set to 1 then collect from the beginning
    :type start: int
    :param count: Collect as many search results as defined by this parameter. If set to 0 then collect until end.
    :type count: int
    :param column_names: names of the columns in the pandas.DataFrame
    :type column_names: ``list``
    :returns: output of 'slk_helpers list_search' parsed into a
        pandas.DataFrame with three columns (permissions, size, filename)
    :rtype: ``pandas.DataFrame``
    """
    # we got a list or set with search ids
    if isinstance(search_id, (list, set)):
        # check type of list/set elements
        if not all([isinstance(sid, (int, str)) for sid in search_id]):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'search_id' has to be 'str', 'int' or 'list'/'set' these "
                + "but it is 'list'/'set' which contains at least elements of these other types: "
                + f"'{', '.join([type(sid).__name__ for sid in search_id if not isinstance(sid, str)])}'"
            )
        # call this function recursively
        return pd.concat(
            [
                list_clone_search(
                    sid,
                    only_files=only_files,
                    only_namespaces=only_namespaces,
                    start=start,
                    count=count,
                    column_names=column_names,
                )
                for sid in search_id
            ],
            ignore_index=True,
        )
    elif isinstance(search_id, str):
        # we have a search id as string which needs to be converted to int
        try:
            search_id = int(search_id)
        except ValueError:
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: cannot convert 'search_id' to 'int' (== search id), got: {search_id}"
            )
    elif not isinstance(search_id, int):
        # we have something else which is neither 'int' nor something we caught before
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'search_id' has wrong type; need 'str', 'int' or 'list'/'set' of these "
            + f"but got {type(search_id).__name__}"
        )

    output: subprocess.CompletedProcess = list_clone_search_raw(
        search_id,
        only_files=only_files,
        only_namespaces=only_namespaces,
        start=start,
        count=count,
        return_type=2,
    )
    stdout = output.stdout.decode("utf-8").rstrip()
    stderr = output.stderr.decode("utf-8").rstrip()
    if output.returncode > 0:
        if "ERROR: No resources found for given search id:" in f"{stdout}  {stderr}":
            return pd.DataFrame(columns=column_names)
        else:
            raise PySlkException(f"pyslk.{inspect.stack()[0][3]}: {stdout} {stderr}")
    # df = pd.read_fwf(list_out, col_specs = 'infer',
    #                 width=column_widths, dtype=str, header=None,
    #                 names=column_names, infer_nrows=10000) #infer_nrows
    # df.drop(df.tail(2).index, inplace=True)

    df = _parse_list(
        stdout,
        search_id,
        column_names,
        True,
        True,
        True,
        True,
    )

    return df


def list_clone_file(
    resource_paths: Union[str, Path, list[str], list[Path], set[str], set[Path]],
    print_resource_ids: bool = False,
    print_timestamps_as_seconds_since_1970: bool = False,
    proceed_on_error: bool = True,
    column_names=PYSLK_DEFAULT_LIST_CLONE_COLUMNS,
) -> pd.DataFrame:
    """Return pandas.DataFrame containing results from search id

    Calls :py:meth:`~pyslk.list_clone_file_raw` and parses the return string into
    a :py:meth:`pandas.DataFrame`. Assumes six output columns of 'slk_helpers list_clone_file'
    having the widths 12, 16, 999 (999 => as wide as necessary).


    :param resource_paths: namespace or resource
    :type resource_paths: ``str`` or ``path-like`` or ``list[str]`` or ``list[Path]`` or ``set[str]`` or ``set[Path]``
    :param print_resource_ids: print resource ids instead of file paths
    :type print_resource_ids: `bool`
    :param print_timestamps_as_seconds_since_1970: print timestamps in seconds since 1970
    :type print_timestamps_as_seconds_since_1970: `bool`
    :param proceed_on_error: Proceed listing files even if an error arose
    :type proceed_on_error: `bool`
    :param column_names: names of the columns in the pandas.DataFrame
    :type column_names: ``list``
    :returns: output of 'slk_helpers list_search' parsed into a
        pandas.DataFrame with three columns (permissions, size, filename)
    :rtype: ``pandas.DataFrame``
    """
    # we got a list or set with search ids
    if isinstance(resource_paths, (list, set)):
        # check type of list/set elements
        if not all([isinstance(res, (Path, str)) for res in resource_paths]):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'resource_path' has to be 'str', 'Path' or 'list'/'set' of "
                + "these but it is 'list'/'set' which contains at least elements of these other types: "
                + f"'{', '.join([type(res).__name__ for res in resource_paths if not isinstance(res, str)])}'"
            )
    elif not isinstance(resource_paths, (str, Path)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'resource_path' has to be 'str', 'Path' or 'list'/'set' of "
            + f"these but got {type(resource_paths).__name__}"
        )

    output: subprocess.CompletedProcess = list_clone_file_raw(
        resource_paths,
        print_resource_ids,
        print_timestamps_as_seconds_since_1970,
        proceed_on_error,
        return_type=2,
    )
    stdout = output.stdout.decode("utf-8").rstrip()
    stderr = output.stderr.decode("utf-8").rstrip()
    if output.returncode == 1:
        if proceed_on_error:
            # print warning on what was not found
            warnings.warn("resources were not found:\n" + stderr)
        else:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: resources were not found:\n{stderr}"
            )
    if output.returncode == 2:
        raise PySlkException(f"pyslk.{inspect.stack()[0][3]}: {stdout} {stderr}")
    # df = pd.read_fwf(list_out, col_specs = 'infer',
    #                 width=column_widths, dtype=str, header=None,
    #                 names=column_names, infer_nrows=10000) #infer_nrows
    # df.drop(df.tail(2).index, inplace=True)

    # when print_resource_ids is set, where should be no processing of the
    # resouce (path) be done
    df = _parse_list(
        stdout,
        "/",
        column_names,
        not print_timestamps_as_seconds_since_1970,
        True,
        True,
        True,
        print_resource_ids,
    )

    return df



import inspect
import json
from pathlib import Path
from typing import Union

from ..constants import SEARCH_QUERY_OPERATORS
from ..raw import gen_file_query_raw, gen_search_query_raw
from .resource_extras import is_namespace

__all__ = [
    "gen_file_query",
    "gen_search_query",
    "gen_file_query_as_dict",
]


def gen_file_query(
    resources: Union[str, list[str], set[str], Path, list[Path], set[Path]],
    recursive: bool = False,
    cached_only: bool = False,
    not_cached: bool = False,
    tape_barcodes: Union[list[str], str, None] = None,
) -> str:
    """Generates a search query that searches for the listed resources

    A search query will be generated which connects all elements of
    'resources' with and 'or'. A 'resource' (an element of 'resources')
    might be one of these:

    - a filename with full path (e.g. /arch/bm0146/k204221/INDEX.txt)
    - a filename without full path (e.g. INDEX.txt)
    - a regex describing a filename (e.g. /arch/bm0146/k204221/.*.txt)
    - a namespace (e.g. /arch/bm0146/k204221 or /arch/bm0146/k204221/)

    Details are given in the slk_helpers documentation at https://docs.dkrz.de

    :param resources: list of resources to be searched for
    :type resources: ``str`` or ``list`` or ``set`` or ``Path``
    :param recursive: do recursive search in the namespaces
    :type recursive: ``bool``
    :param cached_only: do recursive search in the namespaces
    :type cached_only: ``bool``
    :param not_cached: do recursive search in the namespaces
    :type not_cached: ``bool``
    :param tape_barcodes: do recursive search in the namespaces
    :type tape_barcodes: ``list[str]``
    :returns: generated search query
    :rtype: ``str``
    """
    if isinstance(resources, (str, Path)):
        if is_namespace(resources) and not recursive:
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: when 'resources' points to a namespace, 'recursive' has to be set "
                + "to 'True'"
            )
    if isinstance(resources, (list, set)):
        if any([is_namespace(r) for r in resources]) and not recursive:
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: when 'resources' points to at least one namespace, 'recursive' has to "
                + "be set to 'True'"
            )

    output = gen_file_query_raw(
        resources, recursive, True, cached_only, not_cached, tape_barcodes
    )
    return output


def gen_search_query(
    key_value_pairs: Union[str, list[str], set[str]],
    recursive: bool = False,
    search_query: Union[str, None] = None,
) -> str:
    """Generates a search query that searches for the listed resources

    A search query will be generated which connects all elements of
    'resources' with and 'or'. A 'resource' (an element of 'resources')
    might be one of these:

    - a filename with full path (e.g. /arch/bm0146/k204221/INDEX.txt)
    - a filename without full path (e.g. INDEX.txt)
    - a regex describing a filename (e.g. /arch/bm0146/k204221/.*.txt)
    - a namespace (e.g. /arch/bm0146/k204221 or /arch/bm0146/k204221/)

    Details are given in the slk_helpers documentation at https://docs.dkrz.de

    :param key_value_pairs: list of key-value pairs connected via an operator
    :type key_value_pairs: ``str`` or ``list`` or ``set``
    :param recursive: do recursive search in the namespaces
    :type recursive: ``bool``
    :param search_query: an existing search query to be extended
    :type search_query: ``str``
    :returns: generated search query
    :rtype: ``str``
    """
    # type checking and check key-value-pairs for completeness
    if isinstance(key_value_pairs, str):
        if not any([o in key_value_pairs for o in SEARCH_QUERY_OPERATORS]):
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'key_value_pairs' has to be a triple of key-operator-value "
                + f"where the operator has one of these values: {', '.join(SEARCH_QUERY_OPERATORS)}"
            )
    elif isinstance(key_value_pairs, (list, set)):
        if not all([isinstance(kvp, str) for kvp in key_value_pairs]):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'key_value_pairs' has to be 'str' or 'list' of 'str' but it"
                + "is 'list' which contains at least elements of these other types: "
                + f"'{', '.join([type(kvp).__name__ for kvp in key_value_pairs if not isinstance(kvp, str)])}'"
            )
        bad_key_values: list[str] = [
            kvp
            for kvp in key_value_pairs
            if not any([o in kvp for o in SEARCH_QUERY_OPERATORS])
        ]
        if len(bad_key_values) > 0:
            raise ValueError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'key_value_pairs' has to be a triple of key-operator-value "
                + f"where the operator has one of these values: {', '.join(SEARCH_QUERY_OPERATORS)}; this triples "
                + f"have no matching operator: {', '.join(bad_key_values)}"
            )
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'key_value_pairs' has to be 'str' or 'list' of 'str' but is "
            + f"'{type(key_value_pairs).__name__}'"
        )
    if not isinstance(recursive, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'recursive' has to be 'bool' but is "
            + f"'{type(search_query).__name__}'"
        )
    if search_query is not None:
        if not isinstance(search_query, str):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: argument 'search_query' has to be 'str' but is "
                + f"'{type(search_query).__name__}'"
            )

    output: str = gen_search_query_raw(key_value_pairs, recursive, search_query)
    return output


def gen_file_query_as_dict(
    resources: Union[str, list[str], set[str], Path, list[Path], set[Path]],
    recursive: bool = False,
    cached_only: bool = False,
    not_cached: bool = False,
    tape_barcodes: Union[list[str], str, None] = None,
) -> dict:
    """Generates a search query that searches for the listed resources

    A search query will be generated which connects all elements of
    'resources' with and 'or'. A 'resource' (an element of 'resources')
    might be one of these:

    - a filename with full path (e.g. /arch/bm0146/k204221/INDEX.txt)
    - a filename without full path (e.g. INDEX.txt)
    - a regex describing a filename (e.g. /arch/bm0146/k204221/.*.txt)
    - a namespace (e.g. /arch/bm0146/k204221 or /arch/bm0146/k204221/)

    Details are given in the slk_helpers documentation at https://docs.dkrz.de

    :param resources: list of resources to be searched for
    :type resources: ``str`` or ``list`` or ``set`` or ``Path``
    :param recursive: do recursive search in the namespaces
    :type recursive: ``bool``
    :param cached_only: do recursive search in the namespaces
    :type cached_only: ``bool``
    :param not_cached: do recursive search in the namespaces
    :type not_cached: ``bool``
    :param tape_barcodes: do recursive search in the namespaces
    :type tape_barcodes: ``list[str]``
    :returns: generated search query
    :rtype: ``dict``
    """
    return json.loads(
        gen_file_query(
            resources=resources,
            recursive=recursive,
            cached_only=cached_only,
            not_cached=not_cached,
            tape_barcodes=tape_barcodes,
        )
    )
